import datetime
import logging
import time

import numpy as np
import torch
import torch.distributed as dist
import torch.nn.functional as F
from einops import rearrange

from models.criterions import get_sim
from utils.basic_utils import MetricLogger
from utils.distributed import get_rank, get_world_size

logger = logging.getLogger(__name__)


def extract_text_feats(texts, max_txt_l, tokenizer, model, device, return_ids=False):
    num_text = len(texts)
    text_bs = 256
    text_feats = []
    text_atts = []
    
    if return_ids:
        text_ids = []

    for i in range(0, num_text, text_bs):
        text = texts[i : min(num_text, i + text_bs)]
        text_input = tokenizer(
            text,
            padding="max_length",
            truncation=True,
            max_length=max_txt_l,
            return_tensors="pt",
        ).to(device) # NOTE not need to cast

        text_feat = model.encode_text(text_input)[0]
        text_feats.append(text_feat)
        text_atts.append(text_input.attention_mask)
        if return_ids:
            text_ids.append(text_input.input_ids)

    text_feats = torch.cat(text_feats, dim=0)
    text_atts = torch.cat(text_atts, dim=0)
    if return_ids:
        text_ids = torch.cat(text_ids, dim=0)
        return text_feats, text_atts, text_ids
    else:
        return text_feats, text_atts

def extract_vision_feats(data_loader, model, device, config):
    if config.use_half_precision: 
        if config.get('use_bf16', False):
            cast_dtype = torch.bfloat16
        else:
            cast_dtype = torch.float16
    else:
        cast_dtype = None

    image_feats_all = []
    pooled_image_feats_all = []
    metric_logger = MetricLogger(delimiter="  ")
    header = "extracting image feats"
    iterator = metric_logger.log_every(data_loader, 100, header)
    for image, img_id in iterator:
        image = image.to(device, dtype=cast_dtype, non_blocking=True)
        image_feat, pooled_image_feat = model.encode_vision(image, test=True)
        if len(pooled_image_feat.shape) == 2:
            pooled_image_feat = pooled_image_feat.unsqueeze(1)  # make av_fusion happy
        if config.evaluation.eval_frame_ensemble == "concat": 
            if len(image_feat.shape) == 4:
                image_feat = rearrange(image_feat, "b t l c -> b (t l) c").contiguous()
            image_feat = image_feat.unsqueeze(1)  # (bsz, 1, #frm*L, d)
        else:
            assert config.video_input.num_frames == 1, "only support single-frame"
            assert config.evaluation.eval_frame_ensemble in ["mean", "max", "lse"]
        if config.evaluation.eval_offload:
            image_feats_all.append(image_feat.cpu())
            pooled_image_feats_all.append(pooled_image_feat.cpu())
        else:
            image_feats_all.append(image_feat)
            pooled_image_feats_all.append(pooled_image_feat)

    image_feats_all = torch.cat(image_feats_all, dim=0)
    pooled_image_feats_all = torch.cat(pooled_image_feats_all, dim=0)

    return image_feats_all, pooled_image_feats_all

def extract_audio_feats(data_loader, model, device, config):
    if config.use_half_precision: 
        if config.get('use_bf16', False):
            cast_dtype = torch.bfloat16
        else:
            cast_dtype = torch.float16
    else:
        cast_dtype = None

    audio_feats_all = []
    pooled_audio_feats_all = []
    metric_logger = MetricLogger(delimiter="  ")
    header = "extracting audio feats"
    iterator = metric_logger.log_every(data_loader, 100, header)
    for audio, _ in iterator:
        audio = audio.to(device, dtype=cast_dtype, non_blocking=True)
        audio_feat, pooled_audio_feat = model.encode_audio(audio, test=True)
        audio_feat = audio_feat.unsqueeze(1)  # make deep_fusion happy
        pooled_audio_feat = pooled_audio_feat.unsqueeze(1)
        if config.evaluation.eval_offload:
            audio_feats_all.append(audio_feat.cpu())
            pooled_audio_feats_all.append(pooled_audio_feat.cpu())
        else:
            audio_feats_all.append(audio_feat)
            pooled_audio_feats_all.append(pooled_audio_feat)

    audio_feats_all = torch.cat(audio_feats_all, dim=0)

    pooled_audio_feats_all = torch.cat(pooled_audio_feats_all, dim=0)
    return audio_feats_all, pooled_audio_feats_all

def extract_audio_vision_feats(data_loader, model, device, config):
    if config.use_half_precision: 
        if config.get('use_bf16', False):
            cast_dtype = torch.bfloat16
        else:
            cast_dtype = torch.float16
    else:
        cast_dtype = None

    audio_feats_all = []
    pooled_audio_feats_all = []
    image_feats_all = []
    pooled_image_feats_all = []
    metric_logger = MetricLogger(delimiter="  ")
    header = "extracting audio and vision feats"
    iterator = metric_logger.log_every(data_loader, 100, header)
    for media, _ in iterator:
        audio = media[0]
        image = media[1]
        audio = audio.to(device, dtype=cast_dtype, non_blocking=True)
        image = image.to(device, dtype=cast_dtype, non_blocking=True)
        audio_feat, pooled_audio_feat = model.encode_audio(audio, test=True)
        audio_feat = audio_feat.unsqueeze(1)  # make deep_fusion happy
        pooled_audio_feat = pooled_audio_feat.unsqueeze(1)
        image_feat, pooled_image_feat = model.encode_vision(image, test=True)
        if len(pooled_image_feat.shape) == 2:
            pooled_image_feat = pooled_image_feat.unsqueeze(1)  # make av_fusion happy
        if config.evaluation.eval_frame_ensemble == "concat":
            if len(image_feat.shape) == 4:
                image_feat = rearrange(image_feat, "b t l c -> b (t l) c").contiguous()
            image_feat = image_feat.unsqueeze(1)  # (bsz, 1, #frm*L, d)
        else:
            assert config.video_input.num_frames == 1, "only support single-frame"
            assert config.evaluation.eval_frame_ensemble in ["mean", "max", "lse"]
        if config.evaluation.eval_offload:
            audio_feats_all.append(audio_feat.cpu())
            pooled_audio_feats_all.append(pooled_audio_feat.cpu())
            image_feats_all.append(image_feat.cpu())
            pooled_image_feats_all.append(pooled_image_feat.cpu())
        else:
            audio_feats_all.append(audio_feat)
            pooled_audio_feats_all.append(pooled_audio_feat)
            image_feats_all.append(image_feat)
            pooled_image_feats_all.append(pooled_image_feat)

    audio_feats_all = torch.cat(audio_feats_all, dim=0)
    pooled_audio_feats_all = torch.cat(pooled_audio_feats_all, dim=0)
    image_feats_all = torch.cat(image_feats_all, dim=0)
    pooled_image_feats_all = torch.cat(pooled_image_feats_all, dim=0)

    return audio_feats_all, pooled_audio_feats_all, image_feats_all, pooled_image_feats_all


@torch.no_grad()
def evaluation_wrapper(model, data_loader, tokenizer, device, config, prefix=""):
    amp_eval_enabled = config.use_half_precision and not (hasattr(config, "deepspeed") and config.deepspeed.enable)
    logger.info(f"Begin to eval, model_without_ddp.dtype={model.dtype if hasattr(model, 'dtype')  else None}, amp_eval_enabled={amp_eval_enabled}, dtype={torch.bfloat16 if config.get('use_bf16', False) else torch.float16}")
    with torch.cuda.amp.autocast(enabled=amp_eval_enabled, dtype=torch.bfloat16 if config.get('use_bf16', False) else torch.float16):
        i2t_match, t2i_match = None, None
        if "qformer" in config.model.model_cls.lower():
            i2t_match, t2i_match, i2t_sim, t2i_sim, i2t_dsl, t2i_dsl = evaluation_qformer(
                model, data_loader, tokenizer, device, config
            )
        elif "blip" in config.model.model_cls.lower():
            raise NotImplementedError
        elif "clip" in config.model.model_cls.lower() or 'coca' in config.model.model_cls.lower():
            # raise NotImplementedError
            i2t_sim, t2i_sim, i2t_dsl, t2i_dsl = evaluation_clip(
                model, data_loader, tokenizer, device, config
            )
        else:
            i2t_match, t2i_match, i2t_sim, t2i_sim, i2t_dsl, t2i_dsl = evaluation(
                model, data_loader, tokenizer, device, config
            )

        if hasattr(data_loader.dataset, "num_prompts"):
            np = data_loader.dataset.num_prompts
            logger.info(f"Using {np} prompts, we need reshape and mean!!!")
            nt = len(data_loader.dataset.text) // np
            if i2t_match is not None:
                i2t_match = i2t_match.reshape((i2t_match.shape[0], nt, np)).mean(axis=-1)
                t2i_match = t2i_match.reshape((nt, np, t2i_match.shape[1])).mean(axis=1)
            i2t_sim = i2t_sim.reshape((i2t_sim.shape[0], nt, np)).mean(axis=-1)
            t2i_sim = t2i_sim.reshape((nt, np, t2i_sim.shape[1])).mean(axis=1)
            i2t_dsl = i2t_dsl.reshape((i2t_dsl.shape[0], nt, np)).mean(axis=-1)
            t2i_dsl = t2i_dsl.reshape((nt, np, t2i_dsl.shape[1])).mean(axis=1)

    score_pairs = [
        (prefix + "_sim", i2t_sim, t2i_sim),
        (prefix + "_dsl", i2t_dsl, t2i_dsl),
    ]
    if i2t_match is not None:
        if config.evaluation.get('use_dsl_for_match', False):
            score_pairs.append((prefix + "_match (use_dsl)", i2t_match, t2i_match))
        else:
            score_pairs.append((prefix + "_match", i2t_match, t2i_match))

    res = dict()
    for name, i2t, t2i in score_pairs:
        if i2t is not None:
            txt2img_ids = data_loader.dataset.txt2img
            img2txt_ids = data_loader.dataset.img2txt
            res[name] = itm_eval(i2t, t2i, txt2img_ids, img2txt_ids)
    return res


@torch.no_grad()
def evaluation(model, data_loader, tokenizer, device, config):
    model.eval()

    use_dsl_for_match = config.evaluation.get('use_dsl_for_match', False)

    metric_logger = MetricLogger(delimiter="  ")
    header = "Evaluation:"
    dtype = torch.half if config.use_half_precision else torch.float
    media_type = data_loader.dataset.media_type
    use_subtitle = hasattr(data_loader.dataset, "use_subtitle") and data_loader.dataset.use_subtitle
    if use_subtitle:
        assert media_type in ["video", "audio_video"], f"Not support media_type: {media_type}."
        assert hasattr(data_loader.dataset, "subtitle") and data_loader.dataset.subtitle is not None, "You don't have subtitle to use."

    logger.info(f"Start evaluation for media_type={media_type}")
    assert media_type in ['audio', 'video', 'audio_video'], f"Not implement evaluation of {media_type}"

    logger.info("Computing dual encoder features...")
    start_time = time.time()

    # this computes all features in each GPU
    texts = data_loader.dataset.text
    # max_txt_l of eval depends on data_cofig
    max_txt_l = data_loader.dataset.max_txt_l 

    text_feats, text_atts = extract_text_feats(
        texts, max_txt_l, tokenizer, model, device
    )  # (bsz, Lt, d), (bsz, Lt)

    if use_subtitle:
        subtitle_feats, _ = extract_text_feats(
            data_loader.dataset.subtitle, max_txt_l, tokenizer, model, device
        ) # (bsz, Lt, d), (bsz, Lt)
        subtitle_proj = model.text_proj(subtitle_feats[:, 0]).unsqueeze(1)
        subtitle_feats = subtitle_feats.unsqueeze(1) 
        
    if media_type == 'video':
        image_feats, pooled_image_feats = extract_vision_feats(
            data_loader, model, device, config
        )  # (bsz, 1, #frm*Li, d) or (bsz, #frm, Li, d), (bsz, #frm, d)
        logger.info("Finished vision feature extraction")
        logger.info("Computing ITC scores [dot-product]")
        if config.evaluation.eval_offload:
            # image_feats = image_feats.to(device, non_blocking=True) image_feats will cause OOM!!!
            pooled_image_feats = pooled_image_feats.to(device, non_blocking=True)

        if use_subtitle:
            # print(subtitle_proj.shape, pooled_image_feats.shape)
            i2t_scores, t2i_scores = get_sim(
                model.vs_fusion(torch.concat([subtitle_proj, model.vision_proj(pooled_image_feats)], dim=-1)), model.text_proj(text_feats[:, 0])
            )
        else:
            i2t_scores, t2i_scores = get_sim(
                model.vision_proj(pooled_image_feats), model.text_proj(text_feats[:, 0])
            )

        if use_dsl_for_match:
            logger.info("use_dsl_for_match!!!")
            old_i2t_scores, old_t2i_scores = i2t_scores, t2i_scores
            i2t_scores = old_i2t_scores * old_i2t_scores.softmax(dim=0)
            t2i_scores = old_i2t_scores.T * old_i2t_scores.T.softmax(dim=0)

        num_medias = len(data_loader.dataset.image)
        
        # pooled_media_feats = pooled_image_feats
        if use_subtitle:
            media_feats = torch.concat([subtitle_feats, image_feats], dim=-2)
            if hasattr(model, "vstm_head"):
                match_head = model.vstm_head
            else:
                match_head = None
        else:
            media_feats = image_feats
            if hasattr(model, "itm_head"):
                match_head = model.itm_head
            else:
                match_head = None

    elif media_type == 'audio':
        audio_feats, pooled_audio_feats = extract_audio_feats(
            data_loader, model, device, config
        )
        logger.info("Finished audio feature extraction")
        logger.info("Computing ITC scores [dot-product]")
        if config.evaluation.eval_offload:
            pooled_audio_feats = pooled_audio_feats.to(device, non_blocking=True)

        i2t_scores, t2i_scores = get_sim(
            model.audio_proj(pooled_audio_feats), model.text_proj(text_feats[:, 0])
        )

        num_medias = len(data_loader.dataset.audio)
        media_feats = audio_feats
        # pooled_media_feats = pooled_audio_feats
        if hasattr(model, "atm_head"):
            match_head = model.atm_head
        else:
            match_head = None

    elif media_type == 'audio_video':
        audio_feats, pooled_audio_feats, image_feats, pooled_image_feats = extract_audio_vision_feats(
            data_loader, model, device, config
        )
        logger.info("Finished audio and vision feature extraction")

        logger.info("Computing ITC scores [dot-product]")
        if config.evaluation.eval_offload:
            pooled_audio_feats = pooled_audio_feats.to(device, non_blocking=True)
            pooled_image_feats = pooled_image_feats.to(device, non_blocking=True)

        if use_subtitle:
            i2t_scores, t2i_scores = get_sim(
                model.avs_fusion(torch.concat([model.audio_proj(pooled_audio_feats), subtitle_proj, model.vision_proj(pooled_image_feats)], dim=-1)), model.text_proj(text_feats[:, 0])
            )
        else:
            i2t_scores, t2i_scores = get_sim(
                model.av_fusion(torch.concat([model.audio_proj(pooled_audio_feats), model.vision_proj(pooled_image_feats)], dim=-1)), model.text_proj(text_feats[:, 0])
            )

        num_medias = len(data_loader.dataset.image)
        if use_subtitle:
            media_feats = torch.concat([audio_feats, subtitle_feats, image_feats], dim=-2)
            # pooled_media_feats = pooled_audio_feats
            if hasattr(model, "avstm_head"):
                match_head = model.avstm_head
            else:
                match_head = None
        else:
            media_feats = torch.concat([audio_feats, image_feats], dim=-2)
            # pooled_media_feats = pooled_audio_feats
            if hasattr(model, "avtm_head"):
                match_head = model.avtm_head
            else:
                match_head = None
    else:
        raise NotImplementedError(media_type)
    
    logger.info("Computing ITC scores [dot-product], done!")
    
    if match_head is not None:
        i2t_scores_x = torch.full((num_medias, len(texts)), -100.0).to(
            device, torch.float, non_blocking=True
        )

        # computes only part of the scores at each GPU, gather at the end
        logger.info("Rerank dual-encoder results with cross-encoder...")
        num_tasks = get_world_size()
        rank = get_rank()
        # only uses the part associated with the raw eval set
        # compute media2text #
        step = num_medias // num_tasks + 1
        start = rank * step
        end = min(num_medias, start + step)

        text_encoder = model.get_text_encoder()
        iterator = metric_logger.log_every(i2t_scores[start:end], 100, header)
        logger.info(f"i2t_scores.shape {i2t_scores[start:end].shape}")

        # generate score for each clip, and aggregate all clip scores for a video
        n_clip_per_video = (
            media_feats.shape[1] if not config.deep_fusion else media_feats[0].shape[1]
        )

        assert not config.deep_fusion and n_clip_per_video == 1, f"Not implemented for config.deep_fusion={config.deep_fusion} n_clip_per_video={n_clip_per_video}"

        logger.info(
            f"n_clip_per_video={n_clip_per_video}, with eval_frame_ensemble={config.evaluation.eval_frame_ensemble}"
        )

        for i, sims in enumerate(iterator):
            k = min(len(sims), config.evaluation.k_test)
            topk_sim, topk_idx = sims.topk(k=k, dim=0)

            clip_scores = []
            for clip_idx in range(n_clip_per_video):
                if config.deep_fusion:
                    encoder_output = [
                        feat[start + i, clip_idx].to(device, non_blocking=True)
                        if config.evaluation.eval_offload
                        else feat[start + i, clip_idx]
                        for feat in media_feats
                    ]

                else:
                    encoder_output = (
                        media_feats[start + i, clip_idx].to(device, non_blocking=True)
                        if config.evaluation.eval_offload
                        else media_feats[start + i, clip_idx]
                    )  # (#frm*Li, d)

                # new
                bs = 32
                # bs = config.batch_size_test.video
                itm_embeds = []

                if config.deep_fusion:
                    if len(topk_idx) % bs != 0:
                        left = len(topk_idx) % bs
                        left_encoder_output = [feat.repeat(left, 1, 1) for feat in encoder_output]
                        left_encoder_att = [
                            torch.ones(feat.size()[:-1], dtype=torch.long).to(
                                device, non_blocking=True
                            )
                            for feat in left_encoder_output
                        ]
                    encoder_output = [feat.repeat(bs, 1, 1) for feat in encoder_output]
                    encoder_att = [
                        torch.ones(feat.size()[:-1], dtype=torch.long).to(
                            device, non_blocking=True
                        )
                        for feat in encoder_output
                    ]
                else:
                    if len(topk_idx) % bs != 0:
                        left = len(topk_idx) % bs
                        left_encoder_output = encoder_output.repeat(left, 1, 1)  # (k=128, #frm*Li, d)
                        left_encoder_att = torch.ones(left_encoder_output.size()[:-1], dtype=torch.long).to(
                            device, non_blocking=True
                        )
                    encoder_output = encoder_output.repeat(bs, 1, 1)  # (k=128, #frm*Li, d)
                    encoder_att = torch.ones(encoder_output.size()[:-1], dtype=torch.long).to(
                        device, non_blocking=True
                    )

                for j in range(0, len(topk_idx), bs):
                    if j + bs > len(topk_idx):
                        output = text_encoder(
                            encoder_embeds=text_feats[topk_idx[j:]],
                            attention_mask=text_atts[topk_idx[j:]],
                            encoder_hidden_states=left_encoder_output,
                            encoder_attention_mask=left_encoder_att,
                            return_dict=True,
                            mode="fusion",
                        )
                    else:
                        output = text_encoder(
                            encoder_embeds=text_feats[topk_idx[j : j + bs]],
                            attention_mask=text_atts[topk_idx[j : j + bs]],
                            encoder_hidden_states=encoder_output,
                            encoder_attention_mask=encoder_att,
                            return_dict=True,
                            mode="fusion",
                        )
                    batch_itm_embeds = output.last_hidden_state[:, 0]
                    itm_embeds.append(batch_itm_embeds)
                itm_embeds = torch.cat(itm_embeds, dim=0)
                # end new

                score = match_head(itm_embeds)[:, 1]
                clip_scores.append(score)

            if len(clip_scores) == 1:
                score = clip_scores[0]
            else:
                raise NotImplementedError(f"len(clip_scores) == {len(clip_scores)}")

            i2t_scores_x[start + i, topk_idx] = score.to(i2t_scores_x.dtype)

        # compute text2media #
        num_text = len(data_loader.dataset.text)
        t2i_scores_x = torch.full((num_text, num_medias), -100.0).to(
            device, torch.float, non_blocking=True
        )

        step = num_text // num_tasks + 1
        start = rank * step
        end = min(num_text, start + step)

        iterator = metric_logger.log_every(t2i_scores[start:end], 100, header)
        logger.info(f"t2i_scores.shape {t2i_scores[start:end].shape}")
        # generate score for each clip, and aggregate all clip scores for a video
        n_clip_per_video = (
            media_feats.shape[1] if not config.deep_fusion else media_feats[0].shape[1]
        )
        for i, sims in enumerate(iterator):
            k = min(len(sims), config.evaluation.k_test)
            topk_sim, topk_idx = sims.topk(k=k, dim=0)

            clip_scores = []
            for clip_idx in range(n_clip_per_video):
                # new
                bs = 32
                # bs = config.batch_size_test.video
                itm_embeds = []
                for j in range(0, len(topk_idx), bs):

                    if config.deep_fusion:
                        encoder_output = [
                            feat[topk_idx[j : j + bs].cpu(), clip_idx].to(device, non_blocking=True)
                            if config.evaluation.eval_offload
                            else feat[topk_idx[j : j + bs], clip_idx]
                            for feat in media_feats
                        ]
                        encoder_att = [
                            torch.ones(feat.size()[:-1], dtype=torch.long).to(
                                device, non_blocking=True
                            )
                            for feat in encoder_output
                        ]
                    else:
                        encoder_output = (
                            media_feats[topk_idx[j : j + bs].cpu(), clip_idx].to(
                                device, non_blocking=True
                            )
                            if config.evaluation.eval_offload
                            else media_feats[topk_idx[j : j + bs], clip_idx]
                        )
                        encoder_att = torch.ones(encoder_output.size()[:-1], dtype=torch.long).to(
                            device, non_blocking=True
                        )

                    repeat_n = (
                        encoder_output.shape[0]
                        if not config.deep_fusion
                        else encoder_output[0].shape[0]
                    )
                    output = text_encoder(
                        encoder_embeds=text_feats[start + i].repeat(repeat_n, 1, 1),
                        attention_mask=text_atts[start + i].repeat(repeat_n, 1),
                        encoder_hidden_states=encoder_output,
                        encoder_attention_mask=encoder_att,
                        return_dict=True,
                        mode="fusion",
                    )

                    batch_itm_embeds = output.last_hidden_state[:, 0]
                    itm_embeds.append(batch_itm_embeds)

                itm_embeds = torch.cat(itm_embeds, dim=0)
                # end new

                score = match_head(itm_embeds)[:, 1]
                clip_scores.append(score)

            if len(clip_scores) == 1:
                score = clip_scores[0]
            else:
                raise NotImplementedError(f"len(clip_scores) == {len(clip_scores)}")

            t2i_scores_x[start + i, topk_idx] = score.to(t2i_scores_x.dtype)

        logger.info("Compute over!!!")
        if config.distributed:
            logger.info("Gather across GPUs!!!")
            # gather across GPUs
            dist.barrier()
            logger.info("dist.barrier()!!!")
            dist.all_reduce(i2t_scores_x, op=dist.ReduceOp.SUM)
            logger.info("dist.all_reduce(i2t_scores_x, op=dist.ReduceOp.SUM) over!!!")
            dist.all_reduce(t2i_scores_x, op=dist.ReduceOp.SUM)
            logger.info("dist.all_reduce(t2i_scores_x, op=dist.ReduceOp.SUM) over!!!")
        
        if use_dsl_for_match:
            i2t_scores_dsl = i2t_scores
            i2t_scores_dsl_T = t2i_scores
            i2t_scores = old_i2t_scores
            t2i_scores = old_t2i_scores
        else:
            i2t_scores_dsl = i2t_scores.float() * i2t_scores.float().softmax(dim=0)
            i2t_scores_dsl_T = i2t_scores.float().T * i2t_scores.float().T.softmax(dim=0)
    else:
        i2t_scores_dsl = i2t_scores.float() * i2t_scores.float().softmax(dim=0)
        i2t_scores_dsl_T = i2t_scores.float().T * i2t_scores.float().T.softmax(dim=0)

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    logger.info(f"Evaluation time {total_time_str}")


    if match_head is not None:
        return (
            i2t_scores_x.softmax(dim=1).cpu().float().numpy(),
            t2i_scores_x.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.T.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl_T.softmax(dim=1).cpu().float().numpy()
        )
    else:
        return (
            None,
            None,
            i2t_scores.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.T.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl_T.softmax(dim=1).cpu().float().numpy()
        )
    

@torch.no_grad()
def evaluation_simple(model, data_loader, tokenizer, device, config):
    model.eval()

    metric_logger = MetricLogger(delimiter="  ")
    header = "Evaluation:"
    media_type = data_loader.dataset.media_type

    logger.info(f"Start evaluation for media_type={media_type}")
    assert media_type in ['video'], f"Not implement evaluation of {media_type}"

    logger.info("Computing dual encoder features...")
    start_time = time.time()

    # this computes all features in each GPU
    texts = data_loader.dataset.text
    # max_txt_l of eval depends on data_cofig
    max_txt_l = data_loader.dataset.max_txt_l 

    text_feats, text_atts = extract_text_feats(
        texts, max_txt_l, tokenizer, model, device
    )  # (bsz, Lt, d), (bsz, Lt)

      
    if media_type == 'video':
        image_feats, pooled_image_feats = extract_vision_feats(
            data_loader, model, device, config
        )  # (bsz, 1, #frm*Li, d) or (bsz, #frm, Li, d), (bsz, #frm, d)
        logger.info("Finished vision feature extraction")
        logger.info("Computing ITC scores [dot-product]")
        if config.evaluation.eval_offload:
            # image_feats = image_feats.to(device, non_blocking=True) image_feats will cause OOM!!!
            pooled_image_feats = pooled_image_feats.to(device, non_blocking=True)

        i2t_scores, t2i_scores = get_sim(
            model.vision_proj(pooled_image_feats), model.text_proj(text_feats[:, 0])
        )

        num_medias = len(data_loader.dataset.image)
        
        media_feats = image_feats
        if hasattr(model, "itm_head"):
            match_head = model.itm_head
        else:
            match_head = None

    else:
        raise NotImplementedError(media_type)
    
    logger.info("Computing ITC scores [dot-product], done!")
    
    if match_head is not None:
        i2t_scores_x = torch.full((num_medias, len(texts)), -100.0).to(
            device, torch.float, non_blocking=True
        )

        # computes only part of the scores at each GPU, gather at the end
        logger.info("Rerank dual-encoder results with cross-encoder...")
        num_tasks = get_world_size()
        rank = get_rank()
        # only uses the part associated with the raw eval set
        # compute media2text #
        step = num_medias // num_tasks + 1
        start = rank * step
        end = min(num_medias, start + step)

        text_encoder = model.get_text_encoder()
        iterator = metric_logger.log_every(i2t_scores[start:end], 100, header)
        logger.info(f"i2t_scores.shape {i2t_scores[start:end].shape}")

        # generate score for each clip, and aggregate all clip scores for a video
        n_clip_per_video = (
            media_feats.shape[1] if not config.deep_fusion else media_feats[0].shape[1]
        )

        assert not config.deep_fusion and n_clip_per_video == 1, f"Not implemented for config.deep_fusion={config.deep_fusion} n_clip_per_video={n_clip_per_video}"

        logger.info(
            f"n_clip_per_video={n_clip_per_video}, with eval_frame_ensemble={config.evaluation.eval_frame_ensemble}"
        )

        for i, sims in enumerate(iterator):
            k = min(len(sims), config.evaluation.k_test)
            topk_sim, topk_idx = sims.topk(k=k, dim=0)

            clip_scores = []
            for clip_idx in range(n_clip_per_video):
                if config.deep_fusion:
                    encoder_output = [
                        feat[start + i, clip_idx].to(device, non_blocking=True)
                        if config.evaluation.eval_offload
                        else feat[start + i, clip_idx]
                        for feat in media_feats
                    ]

                else:
                    encoder_output = (
                        media_feats[start + i, clip_idx].to(device, non_blocking=True)
                        if config.evaluation.eval_offload
                        else media_feats[start + i, clip_idx]
                    )  # (#frm*Li, d)

                # new
                bs = 32
                # bs = config.batch_size_test.video
                itm_embeds = []

                if config.deep_fusion:
                    if len(topk_idx) % bs != 0:
                        left = len(topk_idx) % bs
                        left_encoder_output = [feat.repeat(left, 1, 1) for feat in encoder_output]
                        left_encoder_att = [
                            torch.ones(feat.size()[:-1], dtype=torch.long).to(
                                device, non_blocking=True
                            )
                            for feat in left_encoder_output
                        ]
                    encoder_output = [feat.repeat(bs, 1, 1) for feat in encoder_output]
                    encoder_att = [
                        torch.ones(feat.size()[:-1], dtype=torch.long).to(
                            device, non_blocking=True
                        )
                        for feat in encoder_output
                    ]
                else:
                    if len(topk_idx) % bs != 0:
                        left = len(topk_idx) % bs
                        left_encoder_output = encoder_output.repeat(left, 1, 1)  # (k=128, #frm*Li, d)
                        left_encoder_att = torch.ones(left_encoder_output.size()[:-1], dtype=torch.long).to(
                            device, non_blocking=True
                        )
                    encoder_output = encoder_output.repeat(bs, 1, 1)  # (k=128, #frm*Li, d)
                    encoder_att = torch.ones(encoder_output.size()[:-1], dtype=torch.long).to(
                        device, non_blocking=True
                    )

                for j in range(0, len(topk_idx), bs):
                    if j + bs > len(topk_idx):
                        output = text_encoder(
                            encoder_embeds=text_feats[topk_idx[j:]],
                            attention_mask=text_atts[topk_idx[j:]],
                            encoder_hidden_states=left_encoder_output,
                            encoder_attention_mask=left_encoder_att,
                            return_dict=True,
                            mode="fusion",
                        )
                    else:
                        output = text_encoder(
                            encoder_embeds=text_feats[topk_idx[j : j + bs]],
                            attention_mask=text_atts[topk_idx[j : j + bs]],
                            encoder_hidden_states=encoder_output,
                            encoder_attention_mask=encoder_att,
                            return_dict=True,
                            mode="fusion",
                        )
                    batch_itm_embeds = output.last_hidden_state[:, 0]
                    itm_embeds.append(batch_itm_embeds)
                itm_embeds = torch.cat(itm_embeds, dim=0)
                # end new

                score = match_head(itm_embeds)[:, 1]
                clip_scores.append(score)

            if len(clip_scores) == 1:
                score = clip_scores[0]
            else:
                raise NotImplementedError(f"len(clip_scores) == {len(clip_scores)}")

            i2t_scores_x[start + i, topk_idx] = score.to(i2t_scores_x.dtype)

        # compute text2media #
        num_text = len(data_loader.dataset.text)
        t2i_scores_x = torch.full((num_text, num_medias), -100.0).to(
            device, torch.float, non_blocking=True
        )

        step = num_text // num_tasks + 1
        start = rank * step
        end = min(num_text, start + step)

        iterator = metric_logger.log_every(t2i_scores[start:end], 100, header)
        logger.info(f"t2i_scores.shape {t2i_scores[start:end].shape}")
        # generate score for each clip, and aggregate all clip scores for a video
        n_clip_per_video = (
            media_feats.shape[1] if not config.deep_fusion else media_feats[0].shape[1]
        )
        for i, sims in enumerate(iterator):
            k = min(len(sims), config.evaluation.k_test)
            topk_sim, topk_idx = sims.topk(k=k, dim=0)

            clip_scores = []
            for clip_idx in range(n_clip_per_video):
                # new
                bs = 32
                # bs = config.batch_size_test.video
                itm_embeds = []
                for j in range(0, len(topk_idx), bs):

                    if config.deep_fusion:
                        encoder_output = [
                            feat[topk_idx[j : j + bs].cpu(), clip_idx].to(device, non_blocking=True)
                            if config.evaluation.eval_offload
                            else feat[topk_idx[j : j + bs], clip_idx]
                            for feat in media_feats
                        ]
                        encoder_att = [
                            torch.ones(feat.size()[:-1], dtype=torch.long).to(
                                device, non_blocking=True
                            )
                            for feat in encoder_output
                        ]
                    else:
                        encoder_output = (
                            media_feats[topk_idx[j : j + bs].cpu(), clip_idx].to(
                                device, non_blocking=True
                            )
                            if config.evaluation.eval_offload
                            else media_feats[topk_idx[j : j + bs], clip_idx]
                        )
                        encoder_att = torch.ones(encoder_output.size()[:-1], dtype=torch.long).to(
                            device, non_blocking=True
                        )

                    repeat_n = (
                        encoder_output.shape[0]
                        if not config.deep_fusion
                        else encoder_output[0].shape[0]
                    )
                    output = text_encoder(
                        encoder_embeds=text_feats[start + i].repeat(repeat_n, 1, 1),
                        attention_mask=text_atts[start + i].repeat(repeat_n, 1),
                        encoder_hidden_states=encoder_output,
                        encoder_attention_mask=encoder_att,
                        return_dict=True,
                        mode="fusion",
                    )

                    batch_itm_embeds = output.last_hidden_state[:, 0]
                    itm_embeds.append(batch_itm_embeds)

                itm_embeds = torch.cat(itm_embeds, dim=0)
                # end new

                score = match_head(itm_embeds)[:, 1]
                clip_scores.append(score)

            if len(clip_scores) == 1:
                score = clip_scores[0]
            else:
                raise NotImplementedError(f"len(clip_scores) == {len(clip_scores)}")

            t2i_scores_x[start + i, topk_idx] = score.to(t2i_scores_x.dtype)

        logger.info("Compute over!!!")
        if config.distributed:
            logger.info("Gather across GPUs!!!")
            # gather across GPUs
            dist.barrier()
            logger.info("dist.barrier()!!!")
            dist.all_reduce(i2t_scores_x, op=dist.ReduceOp.SUM)
            logger.info("dist.all_reduce(i2t_scores_x, op=dist.ReduceOp.SUM) over!!!")
            dist.all_reduce(t2i_scores_x, op=dist.ReduceOp.SUM)
            logger.info("dist.all_reduce(t2i_scores_x, op=dist.ReduceOp.SUM) over!!!")
        
        i2t_scores_dsl = i2t_scores.float() * i2t_scores.float().softmax(dim=0)
        i2t_scores_dsl_T = i2t_scores.float().T * i2t_scores.float().T.softmax(dim=0)
    else:
        i2t_scores_dsl = i2t_scores.float() * i2t_scores.float().softmax(dim=0)
        i2t_scores_dsl_T = i2t_scores.float().T * i2t_scores.float().T.softmax(dim=0)

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    logger.info(f"Evaluation time {total_time_str}")


    if match_head is not None:
        return (
            i2t_scores_x.softmax(dim=1).cpu().float().numpy(),
            t2i_scores_x.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.T.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl_T.softmax(dim=1).cpu().float().numpy()
        )
    else:
        return (
            None,
            None,
            i2t_scores.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.T.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl_T.softmax(dim=1).cpu().float().numpy()
        )
    

@torch.no_grad()
def evaluation_qformer(model, data_loader, tokenizer, device, config):
    model.eval()

    metric_logger = MetricLogger(delimiter="  ")
    header = "Evaluation:"
    dtype = torch.half if config.use_half_precision else torch.float
    media_type = data_loader.dataset.media_type
    logger.info(f"Start evaluation_qformer for media_type={media_type}")
    assert media_type == 'video', f"Not implement evaluation of {media_type}"
    logger.info("Computing dual encoder features...")
    start_time = time.time()

    # this computes all features in each GPU
    texts = data_loader.dataset.text
    # max_txt_l of eval depends on data_cofig
    max_txt_l = data_loader.dataset.max_txt_l 

    text_feats, text_atts, text_ids = extract_text_feats(
        texts, max_txt_l, tokenizer, model, device, return_ids=True
    )  # (bsz, Lt, d), (bsz, Lt)

    if media_type == 'video':
        image_feats, pooled_image_feats = extract_vision_feats(
            data_loader, model, device, config
        )  # (bsz, 1, #frm*Li, d) or (bsz, #frm, Li, d), (bsz, #frm, d)
        logger.info("Finished vision feature extraction")
        logger.info("Computing ITC scores [dot-product]")
        if config.evaluation.eval_offload:
            # image_feats = image_feats.to(device, non_blocking=True) image_feats will cause OOM!!!
            pooled_image_feats = pooled_image_feats.to(device, non_blocking=True)

        if hasattr(model, "q_vision_proj"):
            i2t_scores, t2i_scores = get_sim(
                model.q_vision_proj(pooled_image_feats), model.q_text_proj(text_feats[:, 0])
            )
        else:
            i2t_scores, t2i_scores = get_sim(
                model.vision_proj(pooled_image_feats), model.text_proj(text_feats[:, 0])
            )

        num_medias = len(data_loader.dataset.image)
        
        media_feats = image_feats
        if hasattr(model, "itm_head"):
            match_head = model.itm_head
        elif hasattr(model, "q_itm_head"):
            match_head = model.q_itm_head
        else:
            raise NotImplementedError("you must have a match head in qformer!!!")
    
    logger.info("Computing ITC scores [dot-product], done!")

    if match_head is not None:
        i2t_scores_x = torch.full((num_medias, len(texts)), -100.0).to(
            device, torch.float, non_blocking=True
        )

        # computes only part of the scores at each GPU, gather at the end
        logger.info("Rerank dual-encoder results with cross-encoder...")
        num_tasks = get_world_size()
        rank = get_rank()
        # only uses the part associated with the raw eval set
        # compute image2text #
        step = num_medias // num_tasks + 1
        start = rank * step
        end = min(num_medias, start + step)

        iterator = metric_logger.log_every(i2t_scores[start:end], 100, header)
        logger.info(f"i2t_scores.shape {i2t_scores[start:end].shape}")

        # generate score for each clip, and aggregate all clip scores for a video
        n_clip_per_video = (
            image_feats.shape[1] if not config.deep_fusion else image_feats[0].shape[1]
        )

        assert not config.deep_fusion and n_clip_per_video == 1, f"Not implemented for config.deep_fusion={config.deep_fusion} n_clip_per_video={n_clip_per_video}"

        logger.info(
            f"n_clip_per_video={n_clip_per_video}, with eval_frame_ensemble={config.evaluation.eval_frame_ensemble}"
        )
        for i, sims in enumerate(iterator):
            k = min(len(sims), config.evaluation.k_test)
            topk_sim, topk_idx = sims.topk(k=k, dim=0)

            clip_scores = []
            for clip_idx in range(n_clip_per_video):
                if config.deep_fusion:
                    encoder_output = [
                        feat[start + i, clip_idx].to(device, non_blocking=True)
                        if config.evaluation.eval_offload
                        else feat[start + i, clip_idx]
                        for feat in media_feats
                    ]

                else:
                    encoder_output = (
                        image_feats[start + i, clip_idx].to(device, non_blocking=True)
                        if config.evaluation.eval_offload
                        else image_feats[start + i, clip_idx]
                    )  # (#frm*Li, d)

                # new
                bs = 32
                # bs = config.batch_size_test.video
                itm_embeds = []

                if not config.deep_fusion:  # Create fake list
                    encoder_output = [encoder_output]
                encoder_output = [feat.repeat(bs, 1, 1) for feat in encoder_output]
                encoder_att = [
                    torch.ones(feat.size()[:-1], dtype=torch.long).to(device, non_blocking=True) 
                    for feat in encoder_output
                ]
                
                for j in range(0, len(topk_idx), bs):
                    cur_bs = min(bs, len(topk_idx) - j)
                    encoder_output = [feat[:cur_bs] for feat in encoder_output]
                    encoder_att = [att[:cur_bs] for att in encoder_att]

                    batch_encoder_output = encoder_output if config.deep_fusion else encoder_output[0]
                    batch_encoder_att = encoder_att if config.deep_fusion else encoder_att[0]
                    
                    output = model.vtm_embed(
                        text_ids=text_ids[topk_idx[j:j+bs]],
                        text_atts=text_atts[topk_idx[j:j+bs]],
                        vision_embeds=batch_encoder_output,
                        vision_atts=batch_encoder_att,
                    )


                    itm_embeds.append(output)

                itm_embeds = torch.cat(itm_embeds, dim=0)
                
                score = match_head(itm_embeds)[:, 1]
                clip_scores.append(score)

            if len(clip_scores) == 1:
                score = clip_scores[0]
            else:
                raise NotImplementedError(f"len(clip_scores) == {len(clip_scores)}")

            i2t_scores_x[start + i, topk_idx] = score.to(i2t_scores_x.dtype)

        # compute text2image #
        num_text = len(data_loader.dataset.text)
        t2i_scores_x = torch.full((num_text, len(data_loader.dataset.image)), -100.0).to(
            device, torch.float, non_blocking=True
        )

        step = num_text // num_tasks + 1
        start = rank * step
        end = min(num_text, start + step)

        iterator = metric_logger.log_every(t2i_scores[start:end], 100, header)
        logger.info(f"t2i_scores.shape {t2i_scores[start:end].shape}")
        # generate score for each clip, and aggregate all clip scores for a video
        n_clip_per_video = (
            image_feats.shape[1] if not config.deep_fusion else image_feats[0].shape[1]
        )
        k = config.evaluation.k_test
        logger.info(f"Top-{k} matching")
        for i, sims in enumerate(iterator):
            k = min(len(sims), config.evaluation.k_test)
            topk_sim, topk_idx = sims.topk(k=k, dim=0)

            clip_scores = []
            for clip_idx in range(n_clip_per_video):

                # new
                bs = 32
                # bs = config.batch_size_test.video
                itm_embeds = []
                for j in range(0, len(topk_idx), bs):

                    fake_image_feats = [image_feats] if not config.deep_fusion else image_feats

                    encoder_output = [
                        feat[topk_idx[j : j + bs].cpu(), clip_idx].to(device, non_blocking=True) 
                        if config.evaluation.eval_offload
                        else feat[topk_idx[j : j + bs], clip_idx]
                        for feat in fake_image_feats
                    ]
                    encoder_att = [
                        torch.ones(feat.size()[:-1], dtype=torch.long).to(
                            device, non_blocking=True
                        )
                        for feat in encoder_output
                    ]
                    cur_bs = min(bs, len(topk_idx) - j)

                    batch_encoder_output = encoder_output if config.deep_fusion else encoder_output[0]
                    batch_encoder_att = encoder_att if config.deep_fusion else encoder_att[0]


                    output = model.vtm_embed(
                        text_ids=text_ids[start + i].repeat(cur_bs, 1),
                        text_atts=text_atts[start + i].repeat(cur_bs, 1),
                        vision_embeds=batch_encoder_output,
                        vision_atts=batch_encoder_att,
                    )

                    itm_embeds.append(output)


                itm_embeds = torch.cat(itm_embeds, dim=0)
                # end new

                score = match_head(itm_embeds)[:, 1]
                clip_scores.append(score)

            if len(clip_scores) == 1:
                score = clip_scores[0]
            else:
                raise NotImplementedError(f"len(clip_scores) == {len(clip_scores)}")

            t2i_scores_x[start + i, topk_idx] = score.to(t2i_scores_x.dtype)

        logger.info("Compute over!!!")
        if config.distributed:
            logger.info("Gather across GPUs!!!")
            # gather across GPUs
            dist.barrier()
            logger.info("dist.barrier()!!!")
            dist.all_reduce(i2t_scores_x, op=dist.ReduceOp.SUM)
            logger.info("dist.all_reduce(i2t_scores_x, op=dist.ReduceOp.SUM) over!!!")
            dist.all_reduce(t2i_scores_x, op=dist.ReduceOp.SUM)
            logger.info("dist.all_reduce(t2i_scores_x, op=dist.ReduceOp.SUM) over!!!")

        i2t_scores_dsl = i2t_scores.float() * i2t_scores.float().softmax(dim=0)
        i2t_scores_dsl_T = i2t_scores.float().T * i2t_scores.float().T.softmax(dim=0)

    else:
        i2t_scores_dsl = i2t_scores.float() * i2t_scores.float().softmax(dim=0)
        i2t_scores_dsl_T = i2t_scores.float().T * i2t_scores.float().T.softmax(dim=0)
    
    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    logger.info(f"Evaluation time {total_time_str}")

    i2t_scores_dsl = i2t_scores * i2t_scores.softmax(dim=0)
    i2t_scores_dsl_T = i2t_scores.T * i2t_scores.T.softmax(dim=0)


    if match_head is not None:
        return (
            i2t_scores_x.softmax(dim=1).cpu().float().numpy(),
            t2i_scores_x.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.T.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl_T.softmax(dim=1).cpu().float().numpy()
        )
    else:
        return (
            None,
            None,
            i2t_scores.softmax(dim=1).cpu().float().numpy(),
            i2t_scores.T.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl.softmax(dim=1).cpu().float().numpy(),
            i2t_scores_dsl_T.softmax(dim=1).cpu().float().numpy()
        )


@torch.no_grad()
def evaluation_clip(model, data_loader, tokenizer, device, config):
    model.eval()

    metric_logger = MetricLogger(delimiter="  ")
    header = "Evaluation:"
    dtype = torch.half if config.use_half_precision else torch.float
    media_type = data_loader.dataset.media_type
    logger.info(f"Start evaluation_clip for media_type={media_type}")

    logger.info("Computing dual encoder features...")

    # this computes all features in each GPU
    texts = data_loader.dataset.text
    num_text = len(texts)
    text_bs = 256
    text_feats = []
    for i in range(0, num_text, text_bs):
        text = texts[i : min(num_text, i + text_bs)]
        if "internvideo2" in config.model.model_cls.lower():
            text_feat = model.encode_text(tokenizer(text).to(device))
        else:
            raise NotImplementedError
            text_feat = model.encode_text(text)
        text_feats.append(text_feat.cpu())
    text_feats = torch.cat(text_feats, dim=0)
    logger.info("Finished computing text features")

    media_feats = []
    metric_logger = MetricLogger(delimiter="  ")
    header = f"extracting {media_type} feats!!!"
    iterator = metric_logger.log_every(data_loader, 100, header)
    for media, _ in iterator:
        if media_type in ['image', 'video']:
            media = media.to(device, non_blocking=True)
            media_feat = model.encode_vision(media, test=True)
        elif media_type == 'audio':
            media = media.to(device, non_blocking=True)
            media_feat = model.encode_audio(media, test=True)
        elif media_type == 'audio_video':
            raise NotImplementedError(f"Not implement media_type:{media_type}")
        else:
            raise NotImplementedError(f"Not implement media_type:{media_type}")
        
        media_feats.append(media_feat.cpu())

    media_feats = torch.cat(media_feats, dim=0)
    logger.info("Finished feature extraction")
    logger.info("Computing ITC scores [dot-product]")
    # print(media_feats.dtype, text_feats.dtype)
    # print(media_feats.device, text_feats.device)
    i2t_scores, t2i_scores = get_sim(media_feats.float(), text_feats.float())
    del media_feats, text_feats
    logger.info("Computing ITC scores [dot-product], done!")

    i2t_scores_dsl = i2t_scores * i2t_scores.softmax(dim=0)
    i2t_scores_dsl_T = i2t_scores.T * i2t_scores.T.softmax(dim=0)

    return (
        i2t_scores.cpu().float().numpy(),
        i2t_scores.T.cpu().float().numpy(),
        i2t_scores_dsl.cpu().float().numpy(),
        i2t_scores_dsl_T.cpu().float().numpy(),
    )


@torch.no_grad()
def itm_eval(scores_i2t, scores_t2i, txt2img, img2txt):
    # Images->Text
    ranks = np.zeros(scores_i2t.shape[0])
    for index, score in enumerate(scores_i2t):
        inds = np.argsort(score)[::-1]
        # Score
        gt_txt_ids = img2txt[index]
        if isinstance(gt_txt_ids, int):
            ranks[index] = np.where(inds == gt_txt_ids)[0][0]
        else:
            rank = 1e20
            for i in gt_txt_ids:
                tmp = np.where(inds == i)[0][0]
                if tmp < rank:
                    rank = tmp
            ranks[index] = rank

    # Compute metrics
    tr1 = 100.0 * len(np.where(ranks < 1)[0]) / len(ranks)
    tr5 = 100.0 * len(np.where(ranks < 5)[0]) / len(ranks)
    tr10 = 100.0 * len(np.where(ranks < 10)[0]) / len(ranks)

    # Text->Images
    ranks = np.zeros(scores_t2i.shape[0])

    for index, score in enumerate(scores_t2i):
        inds = np.argsort(score)[::-1]
        gt_img_ids = txt2img[index]
        if isinstance(gt_img_ids, int):
            ranks[index] = np.where(inds == gt_img_ids)[0][0]
        else:  # list, used in the case each caption has multiple GT images
            # Score
            rank = 1e20
            for i in gt_img_ids:
                tmp = np.where(inds == i)[0][0]
                if tmp < rank:
                    rank = tmp
            ranks[index] = rank

    # Compute metrics
    ir1 = 100.0 * len(np.where(ranks < 1)[0]) / len(ranks)
    ir5 = 100.0 * len(np.where(ranks < 5)[0]) / len(ranks)
    ir10 = 100.0 * len(np.where(ranks < 10)[0]) / len(ranks)

    tr_mean = (tr1 + tr5 + tr10) / 3
    ir_mean = (ir1 + ir5 + ir10) / 3
    r_mean = (tr_mean + ir_mean) / 2

    eval_result = {
        "v2t_r1": tr1,
        "v2t_r5": tr5,
        "v2t_r10": tr10,
        "v2t_r_mean": tr_mean,
        "t2v_r1": ir1,
        "t2v_r5": ir5,
        "t2v_r10": ir10,
        "t2v_r_mean": ir_mean,
        "r_mean": r_mean,
    }
    eval_result = {k: round(v, 2) for k, v in eval_result.items()}
    return eval_result





import logging

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy import interpolate
from typing import List

from torch import nn

logger = logging.getLogger(__name__)


def load_temp_embed_with_mismatch(temp_embed_old, temp_embed_new, add_zero=True):
    """
    Add/Remove extra temporal_embeddings as needed.
    https://arxiv.org/abs/2104.00650 shows adding zero paddings works.

    temp_embed_old: (1, num_frames_old, 1, d)
    temp_embed_new: (1, num_frames_new, 1, d)
    add_zero: bool, if True, add zero, else, interpolate trained embeddings.
    """
    # TODO zero pad
    num_frms_new = temp_embed_new.shape[1]
    num_frms_old = temp_embed_old.shape[1]
    logger.info(f"Load temporal_embeddings, lengths: {num_frms_old}-->{num_frms_new}")
    if num_frms_new > num_frms_old:
        if add_zero:
            temp_embed_new[
                :, :num_frms_old
            ] = temp_embed_old  # untrained embeddings are zeros.
        else:
            temp_embed_new = interpolate_temporal_pos_embed(temp_embed_old, num_frms_new)
    elif num_frms_new < num_frms_old:
        temp_embed_new = temp_embed_old[:, :num_frms_new]
    else:  # =
        temp_embed_new = temp_embed_old
    return temp_embed_new


def interpolate_temporal_pos_embed(temp_embed_old, num_frames_new):
    """
    temp_embed_old: (1, num_frames_old, 1, d)
    Returns:
        temp_embed_new: (1, num_frames_new, 1, d)
    """
    temp_embed_old = temp_embed_old.squeeze(2).permute(
        0, 2, 1
    )  # (1, d, num_frames_old)
    temp_embed_new = F.interpolate(
        temp_embed_old, num_frames_new, mode="linear"
    )  # (1, d, num_frames_new)
    temp_embed_new = temp_embed_new.permute(0, 2, 1).unsqueeze(
        2
    )  # (1, num_frames_new, 1, d)
    return temp_embed_new


def interpolate_pos_embed(pos_embed_old, pos_embed_new, num_patches_new):
    """
    Args:
        pos_embed_old: (1, L_old, d), pre-trained
        pos_embed_new: (1, L_new, d), newly initialized, to be replaced by interpolated weights
        num_patches_new:
    """
    # interpolate position embedding
    embedding_size = pos_embed_old.shape[-1]
    num_extra_tokens = pos_embed_new.shape[-2] - num_patches_new
    # height (== width) for the checkpoint position embedding
    orig_size = int((pos_embed_old.shape[-2] - num_extra_tokens) ** 0.5)
    # height (== width) for the new position embedding
    new_size = int(num_patches_new ** 0.5)

    if orig_size != new_size:
        # class_token and dist_token are kept unchanged
        # the extra tokens seems always at the beginning of the position embedding
        extra_tokens = pos_embed_old[:, :num_extra_tokens]
        # only the position tokens are interpolated
        pos_tokens = pos_embed_old[:, num_extra_tokens:]
        pos_tokens = pos_tokens.reshape(
            -1, orig_size, orig_size, embedding_size
        ).permute(0, 3, 1, 2)
        pos_tokens = torch.nn.functional.interpolate(
            pos_tokens, size=(new_size, new_size), mode="bicubic", align_corners=False
        )
        pos_tokens = pos_tokens.permute(0, 2, 3, 1).flatten(1, 2)
        interpolated_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
        logger.info(f"reshape position embedding from {orig_size}**2 to {new_size}**2")
        return interpolated_pos_embed
    else:
        return pos_embed_old


def interpolate_pos_relative_bias_beit(state_dict_old, state_dict_new, patch_shape_new):
    """
    Args:
        state_dict_old: loaded state dict
        state_dict_new: state dict for model with new image size
        patch_shape_new: new model patch_shape
    ref: https://github.com/microsoft/unilm/blob/master/beit/run_class_finetuning.py
    """
    all_keys = list(state_dict_old.keys())
    for key in all_keys:
        if "relative_position_index" in key:
            state_dict_old.pop(key)

        if "relative_position_bias_table" in key:
            rel_pos_bias = state_dict_old[key]
            src_num_pos, num_attn_heads = rel_pos_bias.size()
            dst_num_pos, _ = state_dict_new[key].size()
            dst_patch_shape = patch_shape_new
            if dst_patch_shape[0] != dst_patch_shape[1]:
                raise NotImplementedError()
            num_extra_tokens = dst_num_pos - (dst_patch_shape[0] * 2 - 1) * (
                dst_patch_shape[1] * 2 - 1
            )
            src_size = int((src_num_pos - num_extra_tokens) ** 0.5)
            dst_size = int((dst_num_pos - num_extra_tokens) ** 0.5)
            if src_size != dst_size:
                # logger.info("Position interpolate for %s from %dx%d to %dx%d" % (
                #     key, src_size, src_size, dst_size, dst_size))
                extra_tokens = rel_pos_bias[-num_extra_tokens:, :]
                rel_pos_bias = rel_pos_bias[:-num_extra_tokens, :]

                def geometric_progression(a, r, n):
                    return a * (1.0 - r ** n) / (1.0 - r)

                left, right = 1.01, 1.5
                while right - left > 1e-6:
                    q = (left + right) / 2.0
                    gp = geometric_progression(1, q, src_size // 2)
                    if gp > dst_size // 2:
                        right = q
                    else:
                        left = q

                # if q > 1.090307:
                #     q = 1.090307

                dis = []
                cur = 1
                for i in range(src_size // 2):
                    dis.append(cur)
                    cur += q ** (i + 1)

                r_ids = [-_ for _ in reversed(dis)]

                x = r_ids + [0] + dis
                y = r_ids + [0] + dis

                t = dst_size // 2.0
                dx = np.arange(-t, t + 0.1, 1.0)
                dy = np.arange(-t, t + 0.1, 1.0)

                # logger.info("Original positions = %s" % str(x))
                # logger.info("Target positions = %s" % str(dx))

                all_rel_pos_bias = []

                for i in range(num_attn_heads):
                    z = rel_pos_bias[:, i].view(src_size, src_size).float().numpy()
                    f = interpolate.interp2d(x, y, z, kind="cubic")
                    all_rel_pos_bias.append(
                        torch.Tensor(f(dx, dy))
                        .contiguous()
                        .view(-1, 1)
                        .to(rel_pos_bias.device)
                    )

                rel_pos_bias = torch.cat(all_rel_pos_bias, dim=-1)

                new_rel_pos_bias = torch.cat((rel_pos_bias, extra_tokens), dim=0)
                state_dict_old[key] = new_rel_pos_bias
    return state_dict_old


def tile(x, dim, n_tile):
    init_dim = x.size(dim)
    repeat_idx = [1] * x.dim()
    repeat_idx[dim] = n_tile
    x = x.repeat(*repeat_idx)
    order_index = torch.LongTensor(
        np.concatenate([init_dim * np.arange(n_tile) + i for i in range(init_dim)])
    )
    return torch.index_select(x, dim, order_index.to(x.device))


def mask_logits(target, mask):
    return target * mask + (1 - mask) * (-1e10)


class AllGather(torch.autograd.Function):
    """An autograd function that performs allgather on a tensor."""

    @staticmethod
    def forward(ctx, tensor, args):
        output = [torch.empty_like(tensor) for _ in range(args.world_size)]
        torch.distributed.all_gather(output, tensor)
        ctx.rank = args.rank
        ctx.batch_size = tensor.shape[0]
        return torch.cat(output, dim=0)

    @staticmethod
    def backward(ctx, grad_output):
        return (
            grad_output[ctx.batch_size * ctx.rank : ctx.batch_size * (ctx.rank + 1)],
            None,
        )


allgather_wgrad = AllGather.apply


def tie_encoder_decoder_weights(
    encoder: nn.Module, decoder: nn.Module, base_model_prefix: str, skip_key: str
):
    uninitialized_encoder_weights: List[str] = []
    if decoder.__class__ != encoder.__class__:
        if issubclass(decoder.__class__, encoder.__class__):
            logger.info(
                f"decoder ({decoder.__class__}) and encoder ({encoder.__class__}) are not equal, encoder is decoder's father. In this case make sure that all encoder weights are correctly initialized."
            )
        elif issubclass(encoder.__class__, decoder.__class__):
            logger.info(
                f"decoder ({decoder.__class__}) and encoder ({encoder.__class__}) are not equal, decoder is encoder's father. In this case make sure that all encoder weights are correctly initialized."
            )
        else:
            raise ValueError(f"decoder ({decoder.__class__}) and encoder ({encoder.__class__}) are not equal!!!")

    def tie_encoder_to_decoder_recursively(
        decoder_pointer: nn.Module,
        encoder_pointer: nn.Module,
        module_name: str,
        uninitialized_encoder_weights: List[str],
        skip_key: str,
        depth=0,
    ):
        assert isinstance(decoder_pointer, nn.Module) and isinstance(
            encoder_pointer, nn.Module
        ), f"{decoder_pointer} and {encoder_pointer} have to be of type torch.nn.Module"
        if hasattr(decoder_pointer, "weight") and skip_key not in module_name:
            assert hasattr(encoder_pointer, "weight")
            encoder_pointer.weight = decoder_pointer.weight
            if hasattr(decoder_pointer, "bias"):
                assert hasattr(encoder_pointer, "bias")
                encoder_pointer.bias = decoder_pointer.bias
            logger.info(module_name + " is tied")
            return

        encoder_modules = encoder_pointer._modules
        decoder_modules = decoder_pointer._modules
        if len(decoder_modules) > 0:
            assert (
                len(encoder_modules) > 0
            ), f"Encoder module {encoder_pointer} does not match decoder module {decoder_pointer}"

            all_encoder_weights = set(
                [module_name + "/" + sub_name for sub_name in encoder_modules.keys()]
            )
            encoder_layer_pos = 0
            for name, module in decoder_modules.items():
                if name.isdigit():
                    encoder_name = str(int(name) + encoder_layer_pos)
                    decoder_name = name
                    if not isinstance(
                        decoder_modules[decoder_name],
                        type(encoder_modules[encoder_name]),
                    ) and len(encoder_modules) != len(decoder_modules):
                        # this can happen if the name corresponds to the position in a list module list of layers
                        # in this case the decoder has added a cross-attention that the encoder does not have
                        # thus skip this step and subtract one layer pos from encoder
                        encoder_layer_pos -= 1
                        continue
                elif name not in encoder_modules:
                    continue
                elif depth > 500:
                    raise ValueError(
                        "Max depth of recursive function `tie_encoder_to_decoder` reached. It seems that there is a circular dependency between two or more `nn.Modules` of your model."
                    )
                else:
                    decoder_name = encoder_name = name
                tie_encoder_to_decoder_recursively(
                    decoder_modules[decoder_name],
                    encoder_modules[encoder_name],
                    module_name + "/" + name,
                    uninitialized_encoder_weights,
                    skip_key,
                    depth=depth + 1,
                )
                all_encoder_weights.remove(module_name + "/" + encoder_name)

            uninitialized_encoder_weights += list(all_encoder_weights)

    # tie weights recursively
    tie_encoder_to_decoder_recursively(
        decoder, encoder, base_model_prefix, uninitialized_encoder_weights, skip_key
    )




from functools import partial

import torch.nn as nn
from detectron2 import model_zoo
from detectron2.config import LazyCall as L
from detectron2.data.detection_utils import get_fed_loss_cls_weights
from detectron2.data.samplers import RepeatFactorTrainingSampler
from detectron2.evaluation.lvis_evaluation import LVISEvaluator
from detectron2.modeling import SimpleFeaturePyramid, ViT
from detectron2.modeling.backbone.fpn import LastLevelMaxPool
from detectron2.solver import WarmupParamScheduler
from detectron2_pe.modeling import PEv1_det, get_vit_lr_decay_rate_pev1
from fvcore.common.param_scheduler import MultiStepParamScheduler

from ..COCO.mask_rcnn_vitdet_b_100ep import (  # dataloader,; model,; get_vit_lr_decay_rate,
    lr_multiplier, optimizer, train)
from ..common.coco_loader_lsj import dataloader

train.init_checkpoint = "/checkpoint/vision_encoder/pev1/pev1_rc2_spatial_d2.pt"
train.output_dir = (
    "/checkpoint/vision_encoder/d2_output/lvis/mask_rcnn_PEspatial_G_lvis75ep"
)

model = model_zoo.get_config("common/models/mask_rcnn_vitdet.py").model

model.pixel_mean = [127, 127, 127]
model.pixel_std = [127, 127, 127]
model.input_format = "RGB"

img_size = 1024
embed_dim, depth, num_heads, mlp_ratio, dp = 1536, 50, 16, 8960 / 1536, 0.5
pretrain_img_size, patch_size, window_size = 512, 16, 32
# 12, 24, 36, 49 for global attention
window_block_indexes = (
    list(range(0, 12)) + list(range(13, 24)) + list(range(25, 36)) + list(range(37, 49))
)
# Creates Simple Feature Pyramid from ViT backbone
model.backbone = L(SimpleFeaturePyramid)(
    net=L(PEv1_det)(  # Single-scale ViT backbone
        pretrain_img_size=pretrain_img_size,
        img_size=img_size,
        patch_size=patch_size,
        embed_dim=embed_dim,
        depth=depth,
        num_heads=num_heads,
        drop_path_rate=dp,
        window_size=window_size,
        pt_hw_seq_len=32,
        mlp_ratio=mlp_ratio,
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        window_block_indexes=window_block_indexes,
        residual_block_indexes=[],
        use_rel_pos=True,
        out_feature="last_feat",
        tile_posemb=True,
        use_abs_pos=True,
        pretrain_use_cls_token=False,
        use_act_checkpoint=True,
        init_values=0.1,
    ),
    in_feature="${.net.out_feature}",
    out_channels=256,
    scale_factors=(4.0, 2.0, 1.0, 0.5),
    top_block=L(LastLevelMaxPool)(),
    norm="LN",
    square_pad=img_size,
)

model.roi_heads.num_classes = 1203
model.roi_heads.box_predictor.test_score_thresh = 0.02
model.roi_heads.box_predictor.test_topk_per_image = 300
model.roi_heads.box_predictor.use_sigmoid_ce = True
model.roi_heads.box_predictor.use_fed_loss = True
model.roi_heads.box_predictor.get_fed_loss_cls_weights = (
    lambda: get_fed_loss_cls_weights(dataloader.train.dataset.names, 0.5)
)

train.eval_period = 30000

optimizer.params.lr_factor_func = partial(
    get_vit_lr_decay_rate_pev1, lr_decay_rate=0.9, num_layers=50
)


dataloader.train.dataset.names = "lvis_v1_train"
dataloader.train.sampler = L(RepeatFactorTrainingSampler)(
    repeat_factors=L(
        RepeatFactorTrainingSampler.repeat_factors_from_category_frequency
    )(dataset_dicts="${dataloader.train.dataset}", repeat_thresh=0.001)
)
dataloader.test.dataset.names = "lvis_v1_val"
dataloader.evaluator = L(LVISEvaluator)(
    dataset_name="${..test.dataset.names}",
    max_dets_per_image=300,
    output_dir="${train.output_dir}",
)

dataloader.train.total_batch_size = 64

train.max_iter = 184375


lr_multiplier = L(WarmupParamScheduler)(
    scheduler=L(MultiStepParamScheduler)(
        values=[1.0, 0.1, 0.01],
        milestones=[163889, 177546],
        num_updates=train.max_iter,
    ),
    warmup_length=250 / train.max_iter,
    warmup_factor=0.001,
)

optimizer.params.overrides = {}
optimizer.params.weight_decay_norm = None
optimizer.lr = 5e-5

train.max_iter = train.max_iter * 3 // 4  # 100ep -> 75ep
lr_multiplier.scheduler.milestones = [
    milestone * 3 // 4 for milestone in lr_multiplier.scheduler.milestones
]
lr_multiplier.scheduler.num_updates = train.max_iter



import json

from pycocoevalcap.bleu.bleu import Bleu
from pycocoevalcap.cider.cider import Cider
from pycocoevalcap.meteor.meteor import Meteor
from pycocoevalcap.rouge.rouge import Rouge
from pycocoevalcap.spice.spice import Spice
from pycocoevalcap.tokenizer.ptbtokenizer import PTBTokenizer
# from open_clip import tokenize
from tqdm.auto import tqdm

# from open_clip.tokenizer import _tokenizer
from core.vision_encoder.tokenizer import _tokenizer, tokenize

"""
Code adapted from https://github.com/salaniz/pycocoevalcap/blob/master/eval.py
Thanks to @salaniz for the code!
"""


class COCOEvalCap:
    def __init__(self, results):
        self.evalImgs = []
        self.eval = {}
        self.imgToEval = {}
        self.results = results

    def evaluate(self):
        gts = {}
        res = {}
        for imgId, r in enumerate(self.results):
            gts[imgId] = r["true"]
            res[imgId] = r["gen"]
        # =================================================
        # Set up scorers
        # =================================================
        print("tokenization...")
        tokenizer = PTBTokenizer()
        gts = tokenizer.tokenize(gts)
        res = tokenizer.tokenize(res)

        # =================================================
        # Set up scorers
        # =================================================
        print("setting up scorers...")
        scorers = [
            (Bleu(4), ["Bleu_1", "Bleu_2", "Bleu_3", "Bleu_4"]),
            (Meteor(), "METEOR"),
            (Rouge(), "ROUGE_L"),
            (Cider(), "CIDEr"),
            (Spice(), "SPICE"),
        ]

        # =================================================
        # Compute scores
        # =================================================
        for scorer, method in scorers:
            print("computing %s score..." % (scorer.method()))
            score, scores = scorer.compute_score(gts, res)
            if type(method) == list:
                for sc, scs, m in zip(score, scores, method):
                    self.setEval(sc, m)
                    self.setImgToEvalImgs(scs, gts.keys(), m)
                    print("%s: %0.3f" % (m, sc))
            else:
                self.setEval(score, method)
                self.setImgToEvalImgs(scores, gts.keys(), method)
                print("%s: %0.3f" % (method, score))
        self.setEvalImgs()

    def setEval(self, score, method):
        self.eval[method] = score

    def setImgToEvalImgs(self, scores, imgIds, method):
        for imgId, score in zip(imgIds, scores):
            if not imgId in self.imgToEval:
                self.imgToEval[imgId] = {}
                self.imgToEval[imgId]["image_id"] = imgId
            self.imgToEval[imgId][method] = score

    def setEvalImgs(self):
        self.evalImgs = [eval for imgId, eval in self.imgToEval.items()]


def evaluate(
    model,
    dataloader,
    batch_size,
    device,
    transform,
    train_dataloader=None,
    num_workers=None,
    amp=True,
    verbose=False,
):
    results = []
    image_id = 0
    gt = []
    for idx, (img, captions) in enumerate(tqdm(dataloader)):
        out = model.generate(img.to(device))
        decoded = [
            _tokenizer.decode(i)
            .split("<end_of_text>")[0]
            .replace("<start_of_text>", "")
            .strip()
            for i in out.cpu().numpy()
        ]
        for pred, true in zip(decoded, captions):
            true = [{"caption": t} for t in true]
            pred = [{"caption": pred}]
            results.append({"image_id": image_id, "gen": pred, "true": true})
            image_id += 1
    coco_eval = COCOEvalCap(results)
    coco_eval.evaluate()
    metrics = coco_eval.eval
    # print output evaluation scores
    for metric, score in metrics.items():
        print(f"{metric}: {score:.3f}")
    return metrics



"""Tests for st_mad module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_st_mad_basics():
    from ayase.modules.st_mad import STMADModule
    _test_module_basics(STMADModule, "st_mad")

def test_st_mad_no_reference(image_sample):
    from ayase.modules.st_mad import STMADModule
    m = STMADModule()
    result = m.process(image_sample)
    assert result is image_sample



"""Tests for paranoid_decoder module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_paranoid_decoder_basics():
    from ayase.modules.paranoid_decoder import ParanoidDecoderModule
    _test_module_basics(ParanoidDecoderModule, "paranoid_decoder")

def test_paranoid_decoder_video(video_sample):
    from ayase.modules.paranoid_decoder import ParanoidDecoderModule
    video_sample.quality_metrics = QualityMetrics()
    m = ParanoidDecoderModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



import logging
import os
import sys
from os.path import dirname, join

from utils.config import Config
from utils.distributed import init_distributed_mode, is_main_process
from utils.logger import setup_logger

logger = logging.getLogger(__name__)


def setup_config():
    """Conbine yaml config and command line config with OmegaConf.
    Also converts types, e.g., `'None'` (str) --> `None` (None)
    """
    config = Config.get_config()
    if config.debug:
        config.wandb.enable = False
    return config


def setup_evaluate_config(config):
    """setup evaluation default settings, e.g., disable wandb"""
    assert config.evaluate
    config.wandb.enable = False
    if config.output_dir is None:
        config.output_dir = join(dirname(config.pretrained_path), "eval")
    return config


def setup_output_dir(output_dir, excludes=["code"]):
    """ensure not overwritting an exisiting/non-empty output dir"""
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=False)
    else:
        existing_dirs_files = os.listdir(output_dir)  # list
        remaining = set(existing_dirs_files) - set(excludes)
        remaining = [e for e in remaining if "slurm" not in e]
        remaining = [e for e in remaining if ".out" not in e]
        # assert len(remaining) == 0, f"remaining dirs or files: {remaining}"
        # logger.warn(f"remaining dirs or files: {remaining}")


def setup_main():
    """
    Setup config, logger, output_dir, etc.
    Shared for pretrain and all downstream tasks.
    """
    config = setup_config()
    if hasattr(config, "evaluate") and config.evaluate:
        config = setup_evaluate_config(config)
    init_distributed_mode(config)

    if is_main_process():
        setup_output_dir(config.output_dir, excludes=["code"])
        setup_logger(output=config.output_dir, color=True, name="umt")
        logger.info(f"config: {Config.pretty_text(config)}")
        Config.dump(config, os.path.join(config.output_dir, "config.json"))
    return config



"""Tests for temporal_style module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_temporal_style_basics():
    from ayase.modules.temporal_style import TemporalStyleModule
    _test_module_basics(TemporalStyleModule, "temporal_style")

def test_temporal_style_video(video_sample):
    from ayase.modules.temporal_style import TemporalStyleModule
    video_sample.quality_metrics = QualityMetrics()
    m = TemporalStyleModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""A-FINE (Adaptive Fidelity-Naturalness Evaluator) module.

CVPR 2025. Generalized IQA that handles imperfect references.
Adaptively combines fidelity and naturalness. Both FR and NR variants.
"""

import logging
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class AFINEModule(PipelineModule):
    name = "afine"
    description = "A-FINE adaptive fidelity-naturalness IQA (CVPR 2025)"
    default_config = {"subsample": 4}

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._ml_available = False
        self._model = None

    def setup(self) -> None:
        try:
            import pyiqa
            import torch

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            # Use NR variant by default (no reference needed)
            self._model = pyiqa.create_metric("afine_nr", device=device)
            try:
                self._device = next(self._model.parameters()).device
            except StopIteration:
                self._device = torch.device("cpu")
            self._ml_available = True
            logger.info("A-FINE (NR) model loaded on %s", device)
        except (ImportError, Exception) as e:
            logger.warning("A-FINE unavailable: %s", e)

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        if not self._ml_available:
            return sample
        try:
            import cv2
            import torch

            frames = self._load_frames(sample)
            if not frames:
                return sample

            scores = []
            for frame in frames:
                rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                tensor = (
                    torch.from_numpy(rgb).permute(2, 0, 1).unsqueeze(0).float() / 255.0
                )
                tensor = tensor.to(self._device)
                with torch.no_grad():
                    score = self._model(tensor).item()
                scores.append(score)

            sample.quality_metrics.afine_score = float(np.mean(scores))
        except Exception as e:
            logger.warning("A-FINE processing failed: %s", e)
        return sample

    def _load_frames(self, sample: Sample) -> list:
        import cv2

        subsample = self.config.get("subsample", 4)
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = max(int(cap.get(cv2.CAP_PROP_FRAME_COUNT)), 0)
            indices = list(range(0, total, max(1, total // subsample)))[:subsample]
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            frame = cv2.imread(str(sample.path))
            if frame is not None:
                frames.append(frame)
        return frames



"""Tests for st_lpips module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_st_lpips_basics():
    from ayase.modules.st_lpips import STLPIPSModule
    _test_module_basics(STLPIPSModule, "st_lpips")

def test_st_lpips_video(video_sample):
    from ayase.modules.st_lpips import STLPIPSModule
    video_sample.quality_metrics = QualityMetrics()
    m = STLPIPSModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



VisionEncoders = dict()


TextEncoders = dict()
TextEncoders["bert"] = dict(
    name="bert_base",
    pretrained="bert-base-uncased",
    config="configs/config_bert.json",
    d_model=768,
    fusion_layer=9,
)
TextEncoders["bert_large"] = dict(
    name="bert_large",
    pretrained="google-bert/bert-large-uncased", #"bert-large-uncased",
    config="configs/config_bert_large.json",
    d_model=1024,
    fusion_layer=19,
)
TextEncoders["med_bert"] = dict(
    name="med_bert_base",
    pretrained="bert-base-uncased",
    config="configs/med_config.json",
    d_model=768,
)

TextEncoders["med_bert_large"] = dict(
    name="med_bert_large",
    pretrained="bert-base-uncased", # not a bug, it just follows BLIP.
    config="configs/med_large_config.json",
    d_model=768
)


"""CLI interface for Ayase using Typer."""


import csv
import datetime
import json
import os
import shutil
import subprocess
import sys
from pathlib import Path
import re
from typing import Optional, Iterable, List, Dict, Any

import typer
from rich.console import Console
from rich.table import Table
from typing_extensions import Annotated

from . import __version__
from .config import AyaseConfig
from .pipeline import Pipeline, ModuleRegistry, PipelineModule
from .scanner import scan_dataset, VIDEO_EXTENSIONS, IMAGE_EXTENSIONS
from .models import Sample

app = typer.Typer(
    name="ayase",
    help="Modular media quality metrics toolkit",
    no_args_is_help=True,
)
console = Console()


def _discover_all_modules(config: AyaseConfig) -> None:
    plugin_paths = []
    if config and getattr(config, "pipeline", None):
        plugin_paths = config.pipeline.plugin_folders
    ModuleRegistry.discover_modules(plugin_paths=plugin_paths)


def _print_readiness() -> None:
    readiness = ModuleRegistry.readiness_report()
    if not readiness:
        return
    table = Table(title="Module Readiness")
    table.add_column("Module")
    table.add_column("Status")
    table.add_column("Error")
    for name, info in sorted(readiness.items()):
        status = "READY" if info.get("status") == "ready" else "MISSING"
        error = info.get("error") or ""
        table.add_row(name, status, error)
    console.print(table)


def _select_modules(quick: bool, deep: bool, config: AyaseConfig) -> List[str]:
    _discover_all_modules(config)
    all_modules = list(ModuleRegistry.list_modules().keys())
    if deep:
        return sorted(all_modules)
    if quick:
        return [name for name in ["metadata", "basic_quality"] if name in all_modules]
    if config.pipeline.modules:
        return [name for name in config.pipeline.modules if name in all_modules]
    # Default "balanced" set
    preferred = [
        "metadata",
        "basic_quality",
        "structural",
        "motion",
        "text_detection",
        "watermark_classifier",
        "aesthetic_scoring",
        "video_text_matching",
    ]
    return [name for name in preferred if name in all_modules]


def _parse_pipeline_str(pipeline_str: str, config: AyaseConfig) -> List[PipelineModule]:
    """Parse a pipeline string like 'metadata,motion{sample_rate=10}'."""
    _discover_all_modules(config)
    modules = []

    # Simple regex to split by comma but ignore commas inside curly braces
    parts = re.split(r",(?![^{]*})", pipeline_str)

    for part in parts:
        part = part.strip()
        if not part:
            continue

        params = {}
        if "{" in part and part.endswith("}"):
            name, param_str = part[:-1].split("{", 1)
            name = name.strip()
            # Parse params: key=value
            for pair in param_str.split(","):
                if "=" in pair:
                    k, v = pair.split("=", 1)
                    k = k.strip()
                    v = v.strip()
                    # Try to convert to int/float/bool
                    if v.lower() == "true":
                        v = True
                    elif v.lower() == "false":
                        v = False
                    else:
                        try:
                            if "." in v:
                                v = float(v)
                            else:
                                v = int(v)
                        except ValueError:
                            pass
                    params[k] = v
        else:
            name = part

        module_cls = ModuleRegistry.get_module(name)
        if module_cls:
            try:
                if "models_dir" not in params and config:
                    params["models_dir"] = str(config.general.models_dir)
                modules.append(module_cls(config=params))
            except Exception as e:
                console.print(f"[red]Error initializing module '{name}': {e}[/red]")
        else:
            console.print(f"[yellow]Warning: Module '{name}' not found.[/yellow]")

    return modules


def _process_samples(
    pipeline: Pipeline, samples: Iterable[Sample],
) -> Pipeline:
    for sample in samples:
        processed = pipeline.process_sample(sample)
        pipeline.results[str(processed.path)] = processed
    return pipeline


def _write_markdown_report(pipeline: Pipeline) -> str:
    stats = pipeline.stats
    lines = [
        "# Ayase Report",
        "",
        f"- Total samples: {stats.total_samples}",
        f"- Valid samples: {stats.valid_samples}",
        f"- Invalid samples: {stats.invalid_samples}",
        "",
        "## Issues",
        "",
    ]
    for sample in pipeline.results.values():
        if not sample.validation_issues:
            continue
        lines.append(f"### {sample.path}")
        for issue in sample.validation_issues:
            lines.append(f"- {issue.severity.value.upper()}: {issue.message}")
        lines.append("")
    return "\n".join(lines).strip()


def _export_artifacts(pipeline: Pipeline, config: AyaseConfig, label: str) -> Optional[Path]:
    output_dir = Path(config.output.artifacts_dir) if config and config.output else Path("reports")
    fmt = (config.output.artifacts_format if config and config.output else None) or "json"
    try:
        output_dir.mkdir(parents=True, exist_ok=True)
        timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        output_path = output_dir / f"ayase_{label}_{timestamp}.{fmt}"
        pipeline.export_report(output_path, format=fmt)
        console.print(f"[green]Artifact saved:[/green] {output_path}")
        return output_path
    except Exception as e:
        console.print(f"[yellow]Artifact export failed: {e}[/yellow]")
        return None


def _instantiate_modules(module_names: List[str], config: AyaseConfig) -> List[PipelineModule]:
    modules = []
    for name in module_names:
        module_cls = ModuleRegistry.get_module(name)
        if not module_cls:
            continue
        params = {
            "models_dir": str(config.general.models_dir),
        }
        try:
            modules.append(module_cls(config=params))
        except Exception as e:
            console.print(f"[red]Error initializing module '{name}': {e}[/red]")
    return modules


def version_callback(value: bool) -> None:
    """Print version and exit."""
    if value:
        console.print(f"[bold]Ayase[/bold] version {__version__}")
        raise typer.Exit()


@app.callback()
def main(
    version: Annotated[
        Optional[bool],
        typer.Option("--version", "-V", callback=version_callback, is_eager=True),
    ] = None,
    verbose: Annotated[bool, typer.Option("--verbose", "-v", help="Enable verbose output")] = False,
) -> None:
    """Ayase - Modular media quality metrics toolkit."""
    if verbose:
        import logging
        logging.basicConfig(level=logging.DEBUG, format="%(name)s %(levelname)s: %(message)s")
        console.print("[dim]Verbose mode enabled[/dim]")


@app.command()
def scan(
    dataset_path: Annotated[
        Optional[Path], typer.Argument(help="Path to dataset directory")
    ] = None,
    output: Annotated[
        Optional[Path], typer.Option("--output", "-o", help="Output report file (default: stdout)")
    ] = None,
    format: Annotated[
        str,
        typer.Option("--format", "-f", help="Report format: json|csv|markdown|html"),
    ] = "markdown",
    pipeline: Annotated[
        Optional[str],
        typer.Option(
            "--pipeline",
            "-p",
            help="Explicit pipeline modules (e.g. 'metadata,motion{sample_rate=10}')",
        ),
    ] = None,
    modules_flag: Annotated[
        Optional[str],
        typer.Option(
            "--modules",
            help="Comma-separated module names (e.g. 'metadata,basic_quality,motion')",
        ),
    ] = None,
    jobs: Annotated[
        Optional[int],
        typer.Option("--jobs", "-j", help="Number of parallel jobs"),
    ] = None,
    quick: Annotated[
        bool,
        typer.Option("--quick", help="Quick scan (skip quality metrics)"),
    ] = False,
    deep: Annotated[
        bool,
        typer.Option("--deep", help="Deep scan (all quality metrics, slower)"),
    ] = False,
) -> None:
    """Scan a dataset and generate a quality metrics report."""
    is_quiet = format == "json" and output is None

    if not is_quiet:
        console.print(f"[bold blue]Scanning dataset:[/bold blue] {dataset_path}")
        console.print(f"[bold]Output format:[/bold] {format}")

        if jobs:
            console.print(f"[bold]Parallel jobs:[/bold] {jobs}")
        if quick:
            console.print("[yellow]Quick scan mode enabled[/yellow]")
        if deep:
            console.print("[cyan]Deep scan mode enabled[/cyan]")

    config = AyaseConfig.load()
    if dataset_path is None:
        dataset_path = config.pipeline.dataset_path
    if dataset_path is None:
        console.print("[red]Dataset path is required.[/red]")
        raise typer.Exit(code=1)

    if pipeline:
        modules = _parse_pipeline_str(pipeline, config)
    elif modules_flag:
        _discover_all_modules(config)
        module_names = [n.strip() for n in modules_flag.split(",") if n.strip()]
        modules = _instantiate_modules(module_names, config)
    else:
        module_names = _select_modules(quick, deep, config)
        modules = _instantiate_modules(module_names, config)

    p = Pipeline(modules)
    p.start()
    samples = scan_dataset(dataset_path, include_videos=True, include_images=True)
    _process_samples(p, samples)
    p.stop()
    _export_artifacts(p, config, "scan")

    if format == "markdown":
        report = _write_markdown_report(p)
        if output:
            output.write_text(report, encoding="utf-8")
        else:
            console.print(report)
        return

    if output:
        p.export_report(output, format=format)
    else:
        if format == "json":
            data = {
                "stats": p.stats.model_dump(),
                "samples": [s.model_dump(mode="json") for s in p.results.values()],
            }
            console.print_json(data=data)
        elif format == "csv":
            writer = csv.writer(sys.stdout)
            writer.writerow(["Path", "Valid", "Issues", "Recommendations", "Technical Score"])
            for s in p.results.values():
                issues_str = "; ".join([i.message for i in s.validation_issues])
                recs_str = "; ".join(
                    [i.recommendation for i in s.validation_issues if i.recommendation]
                )
                score = (s.quality_metrics.technical_score if s.quality_metrics else None) or 0.0
                writer.writerow([str(s.path), s.is_valid, issues_str, recs_str, f"{score:.2f}"])
        elif format == "html":
            import tempfile
            with tempfile.NamedTemporaryFile(suffix=".html", delete=False, mode="w") as tmp:
                tmp_path = Path(tmp.name)
            try:
                p.export_report(tmp_path, format="html")
                console.print(tmp_path.read_text(encoding="utf-8"))
            finally:
                tmp_path.unlink(missing_ok=True)
        else:
            console.print(f"[red]Unknown format: {format}[/red]")
            raise typer.Exit(code=1)


@app.command()
def run(
    paths: Annotated[List[Path], typer.Argument(help="Paths to files or directories")],
    pipeline: Annotated[
        str,
        typer.Option(
            "--pipeline",
            "-p",
            help="Pipeline configuration (e.g. 'metadata,motion{sample_rate=10}')",
        ),
    ],
    format: Annotated[
        str,
        typer.Option("--format", "-f", help="Output format: json|csv|markdown"),
    ] = "json",
    output: Annotated[Optional[Path], typer.Option("--output", "-o", help="Output file")] = None,
    recursive: Annotated[
        bool,
        typer.Option("--recursive/--no-recursive", help="Scan directories recursively"),
    ] = True,
) -> None:
    """Run a specific quality assessment pipeline on target paths."""
    config = AyaseConfig.load()
    modules = _parse_pipeline_str(pipeline, config)
    p = Pipeline(modules)
    p.start()

    all_samples = []
    for path in paths:
        if path.is_dir():
            all_samples.extend(
                scan_dataset(path, include_videos=True, include_images=True, recursive=recursive)
            )
        else:
            suffix = path.suffix.lower()
            is_video = suffix in VIDEO_EXTENSIONS
            is_image = suffix in IMAGE_EXTENSIONS
            if not is_video and not is_image:
                console.print(f"[red]Unsupported file type: {path}[/red]")
                continue
            all_samples.append(Sample(path=path, is_video=is_video))

    if not all_samples:
        console.print("[yellow]No valid files found to process.[/yellow]")
        raise typer.Exit(code=0)

    _process_samples(p, all_samples)
    p.stop()
    _export_artifacts(p, config, "run")

    if format == "json":
        data = {
            "stats": p.stats.model_dump(),
            "samples": [s.model_dump(mode="json") for s in p.results.values()],
        }
        if output:
            output.write_text(json.dumps(data, indent=2), encoding="utf-8")
        else:
            console.print_json(data=data)
    elif format == "markdown":
        report = _write_markdown_report(p)
        if output:
            output.write_text(report, encoding="utf-8")
        else:
            console.print(report)
    elif format == "csv":
        if output:
            p.export_report(output, format="csv")
        else:
            writer = csv.writer(sys.stdout)
            writer.writerow(["Path", "Valid", "Issues", "Technical Score"])
            for s in p.results.values():
                issues = "; ".join([i.message for i in s.validation_issues])
                score = (s.quality_metrics.technical_score if s.quality_metrics else None) or 0.0
                writer.writerow([str(s.path), s.is_valid, issues, f"{score:.2f}"])


@app.command()
def filter(
    dataset_path: Annotated[Path, typer.Argument(help="Path to dataset directory")],
    output: Annotated[Path, typer.Option("--output", "-o", help="Output directory")],
    min_score: Annotated[
        Optional[int],
        typer.Option("--min-score", help="Minimum metric score [0-100]"),
    ] = None,
    metric: Annotated[
        str,
        typer.Option("--metric", help="Metric to filter by (default: technical_score)"),
    ] = "technical_score",
    mode: Annotated[
        str,
        typer.Option("--mode", help="Filter mode: symlink|copy|list"),
    ] = "list",
    aspect_ratio: Annotated[
        Optional[str],
        typer.Option("--aspect-ratio", help="Filter by aspect ratio (e.g., 16:9)"),
    ] = None,
    resolution: Annotated[
        Optional[str],
        typer.Option("--resolution", help="Filter by resolution (e.g., 1280x720)"),
    ] = None,
) -> None:
    """Filter dataset based on quality metrics."""
    console.print(f"[bold magenta]Filtering dataset:[/bold magenta] {dataset_path}")
    console.print(f"[bold]Output:[/bold] {output}")
    console.print(f"[bold]Mode:[/bold] {mode}")

    if min_score is not None:
        console.print(f"[bold]Minimum {metric}:[/bold] {min_score}")
    if aspect_ratio:
        console.print(f"[bold]Aspect ratio:[/bold] {aspect_ratio}")
    if resolution:
        console.print(f"[bold]Resolution:[/bold] {resolution}")

    config = AyaseConfig.load()
    module_names = _select_modules(quick=False, deep=False, config=config)
    modules = _instantiate_modules(module_names, config)
    pipeline = Pipeline(modules)
    pipeline.start()

    samples = scan_dataset(dataset_path, include_videos=True, include_images=True)
    _process_samples(pipeline, samples)
    pipeline.stop()

    target_ar = None
    if aspect_ratio:
        try:
            w, h = aspect_ratio.split(":")
            target_ar = float(w) / float(h)
        except Exception:
            console.print("[red]Invalid aspect ratio format; expected W:H[/red]")
            raise typer.Exit(code=1)

    target_res = None
    if resolution:
        try:
            w, h = resolution.lower().split("x")
            target_res = (int(w), int(h))
        except Exception:
            console.print("[red]Invalid resolution format; expected WxH[/red]")
            raise typer.Exit(code=1)

    candidates = []
    for sample in pipeline.results.values():
        if min_score is not None:
            score = 0.0
            if sample.quality_metrics:
                score = getattr(sample.quality_metrics, metric, None) or 0.0
            if score < min_score:
                continue
        if target_ar and sample.aspect_ratio and abs(sample.aspect_ratio - target_ar) > 0.01:
            continue
        if target_res and (sample.width is None or sample.height is None or (sample.width, sample.height) != target_res):
            continue
        candidates.append(sample)

    if mode == "list":
        for sample in candidates:
            console.print(str(sample.path))
        return

    if not output:
        console.print("[red]Output directory required for copy/symlink modes.[/red]")
        raise typer.Exit(code=1)

    output.mkdir(parents=True, exist_ok=True)
    for sample in candidates:
        dest = output / sample.path.name
        if mode == "copy":
            shutil.copy2(sample.path, dest)
        elif mode == "symlink":
            try:
                os.symlink(sample.path, dest)
            except OSError:
                shutil.copy2(sample.path, dest)
        else:
            console.print(f"[red]Unknown mode: {mode}[/red]")
            raise typer.Exit(code=1)


@app.command()
def stats(
    dataset_path: Annotated[Path, typer.Argument(help="Path to dataset directory")],
    format: Annotated[
        str,
        typer.Option("--format", "-f", help="Output format: text|json|html"),
    ] = "text",
    chart: Annotated[
        bool,
        typer.Option("--chart", help="Generate charts (requires HTML format)"),
    ] = False,
) -> None:
    """Generate statistics and distribution analysis."""
    console.print(f"[bold cyan]Generating statistics for:[/bold cyan] {dataset_path}")
    console.print(f"[bold]Format:[/bold] {format}")

    if chart:
        console.print("[bold]Charts enabled[/bold]")

    config = AyaseConfig.load()
    module_names = _select_modules(quick=True, deep=False, config=config)
    modules = _instantiate_modules(module_names, config)
    pipeline = Pipeline(modules)
    pipeline.start()
    samples = scan_dataset(dataset_path, include_videos=True, include_images=False)
    _process_samples(pipeline, samples)
    pipeline.stop()

    if format == "json":
        console.print_json(data=pipeline.stats.model_dump())
        return
    if format == "html":
        import tempfile
        with tempfile.NamedTemporaryFile(suffix=".html", delete=False, mode="w") as tmp:
            tmp_path = Path(tmp.name)
        try:
            pipeline.export_report(tmp_path, format="html")
            console.print(tmp_path.read_text(encoding="utf-8"))
        finally:
            tmp_path.unlink(missing_ok=True)
        return

    stats = pipeline.stats
    console.print(f"Total samples: {stats.total_samples}")
    console.print(f"Valid samples: {stats.valid_samples}")
    console.print(f"Invalid samples: {stats.invalid_samples}")


# Modules subcommand
modules_app = typer.Typer(help="Inspect available pipeline modules and plugins")
app.add_typer(modules_app, name="modules")


@modules_app.command("list")
def modules_list() -> None:
    """List all discovered pipeline modules (built-in + plugins)."""
    config = AyaseConfig.load()
    _discover_all_modules(config)
    all_modules = ModuleRegistry.list_modules()

    if not all_modules:
        console.print("[yellow]No modules discovered.[/yellow]")
        raise typer.Exit(code=0)

    table = Table(title="Available Modules")
    table.add_column("Name", style="cyan")
    table.add_column("Description", style="white")
    for name in sorted(all_modules):
        table.add_row(name, all_modules[name])
    console.print(table)
    console.print(f"\n[dim]{len(all_modules)} module(s) total[/dim]")


@modules_app.command("check")
def modules_check() -> None:
    """Verify all modules can be loaded without errors."""
    config = AyaseConfig.load()
    _discover_all_modules(config)
    all_modules = ModuleRegistry.list_modules()

    if not all_modules:
        console.print("[yellow]No modules discovered.[/yellow]")
        raise typer.Exit(code=0)

    errors = 0
    for name in sorted(all_modules):
        cls = ModuleRegistry.get_module(name)
        try:
            cls()
            console.print(f"  [green]OK[/green]  {name}")
        except Exception as e:
            console.print(f"  [red]FAIL[/red] {name}: {e}")
            errors += 1

    if errors:
        console.print(f"\n[red]{errors} module(s) failed to load.[/red]")
        raise typer.Exit(code=1)
    console.print(f"\n[green]All {len(all_modules)} module(s) loaded successfully.[/green]")


@modules_app.command("docs")
def modules_docs(
    output: Annotated[Optional[Path], typer.Option("--output", "-o", help="Output file (default: METRICS.md)")] = None,
    run_tests: Annotated[bool, typer.Option("--run-tests/--no-tests", help="Run tests and show pass/fail status")] = True,
) -> None:
    """Generate METRICS.md with charts, test status, and version info.

    Single command to regenerate everything:
        ayase modules docs -o METRICS.md --run-tests
    """
    config = AyaseConfig.load()
    _discover_all_modules(config)

    from .metrics_doc import generate_metrics_doc

    if run_tests:
        console.print("[cyan]Running tests to collect status...[/cyan]")
    content = generate_metrics_doc(run_tests=run_tests)

    if output:
        output.write_text(content, encoding="utf-8")
        console.print(f"[green]Written to {output}[/green]")
        # List generated charts
        docs_dir = Path("docs")
        if docs_dir.exists():
            charts = [f for f in docs_dir.iterdir() if f.suffix == ".png"]
            if charts:
                console.print(f"[green]Generated {len(charts)} charts in docs/[/green]")
    else:
        print(content)


@modules_app.command("models")
def modules_models(
    output: Annotated[Optional[Path], typer.Option("--output", "-o", help="Output file (default: stdout)")] = None,
) -> None:
    """Generate MODELS.md catalog of all ML models and weights."""
    config = AyaseConfig.load()
    _discover_all_modules(config)

    from .models_doc import generate_models_doc

    content = generate_models_doc()

    if output:
        output.write_text(content, encoding="utf-8")
        console.print(f"[green]Written to {output}[/green]")
    else:
        print(content)


@modules_app.command("sync-readme")
def modules_sync_readme(
    readme: Annotated[Path, typer.Option("--readme", "-r", help="README.md path")] = Path("README.md"),
) -> None:
    """Update module/field counts in README.md to match reality."""
    import re as _re

    config = AyaseConfig.load()
    _discover_all_modules(config)

    all_modules = ModuleRegistry.list_modules()
    total = len([n for n in all_modules if ModuleRegistry.get_module(n) is not None])

    from .models import QualityMetrics
    n_fields = len(QualityMetrics.model_fields)

    if not readme.exists():
        console.print(f"[red]{readme} not found[/red]")
        raise typer.Exit(code=1)

    text = readme.read_text(encoding="utf-8")
    new_text = _re.sub(
        r"\*\*\d+ modules\*\*,\s*\*\*\d+ quality metrics\*\*",
        f"**{total} modules**, **{n_fields} quality metrics**",
        text,
    )
    if new_text != text:
        readme.write_text(new_text, encoding="utf-8")
        console.print(f"[green]Updated README.md: {total} modules, {n_fields} fields[/green]")
    else:
        console.print(f"[green]README.md already up to date ({total} modules, {n_fields} fields)[/green]")


# Config subcommand
config_app = typer.Typer(help="Manage configuration")
app.add_typer(config_app, name="config")


@config_app.command("init")
def config_init() -> None:
    """Initialize default config."""
    console.print("[bold]Initializing default config...[/bold]")
    config = AyaseConfig.load()
    path = Path("ayase.toml")
    if path.exists():
        console.print("[yellow]ayase.toml already exists; not overwriting.[/yellow]")
        return
    config.save(path)
    console.print(f"[green]Wrote default config to {path}[/green]")


@config_app.command("show")
def config_show() -> None:
    """Show current config."""
    console.print("[bold]Current configuration:[/bold]")
    config = AyaseConfig.load()
    table = Table(title="Configuration")
    table.add_column("Setting", style="cyan")
    table.add_column("Value", style="green")
    for key, value in config.model_dump().items():
        table.add_row(str(key), json.dumps(value, default=str))
    console.print(table)


@config_app.command("edit")
def config_edit() -> None:
    """Edit config in $EDITOR."""
    path = Path("ayase.toml")
    if not path.exists():
        AyaseConfig.load().save(path)
    editor = os.environ.get("EDITOR")
    if not editor:
        editor = "notepad" if os.name == "nt" else "vi"
    console.print(f"[bold]Opening config in editor ({editor})...[/bold]")
    subprocess.run([editor, str(path)])


@config_app.command("validate")
def config_validate() -> None:
    """Validate config file."""
    console.print("[bold green]Validating config file...[/bold green]")
    try:
        AyaseConfig.load()
        console.print("[green]Configuration is valid.[/green]")
    except Exception as e:
        console.print(f"[red]Configuration invalid: {e}[/red]")
        raise typer.Exit(code=1)


@app.command()
def tui() -> None:
    """Launch the Terminal User Interface (TUI)."""
    try:
        from .tui import AyaseApp

        app = AyaseApp()
        app.run()
    except ImportError as e:
        console.print(f"[red]Error launching TUI: {e}[/red]")
        console.print(
            "[yellow]Please ensure 'textual' is installed: pip install ayase[tui][/yellow]"
        )
    except Exception as e:
        console.print(f"[red]Unexpected error: {e}[/red]")


if __name__ == "__main__":
    app()



# Copyright (c) Meta Platforms, Inc. and affiliates.

import itertools
import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
from torch import nn
from torch.distributed._tensor import Replicate, Shard
from torch.distributed.tensor.parallel import (
    ColwiseParallel,
    PrepareModuleInput,
    RowwiseParallel,
    SequenceParallel,
    parallelize_module,
)
from torch.nn.attention.flex_attention import BlockMask, create_block_mask
from xformers.ops import AttentionBias, fmha

from ...core.transformer import (
    BaseTransformer,
    BaseTransformerArgs,
    RMSNorm,
    TiedLinear,
    cross_entropy,
)
from ...core.utils import InitArgs
from ...core.vision_encoder.pe import VisionTransformer as PE_VisionTransformer
from ...core.vision_projector.mlp import MLPProjector

logger = logging.getLogger(__name__)


def create_causal_mask(seqlen, attn_impl, sliding_window):
    if sliding_window is not None and attn_impl == "xformers":
        return fmha.attn_bias.LocalAttentionFromBottomRightMask(
            window_left=sliding_window - 1, window_right=0
        )
    elif attn_impl == "xformers":
        return fmha.attn_bias.LowerTriangularMask()
    elif attn_impl == "sdpa":
        return "causal"
    elif attn_impl == "flex_attention":
        return create_block_mask(causal_mask, None, None, seqlen, seqlen)
    else:
        raise NotImplementedError(
            f"Attention {attn_impl} with {sliding_window} sliding window not implemented"
        )


def attention_flops_per_token(n_layers, seq_len, dim, causal):
    # Formula from https://github.com/Dao-AILab/flash-attention/blob/main/benchmarks/benchmark_flash_attention.py#L27-L30
    return 3.5 * (4 * n_layers * seq_len * dim // (2 if causal else 1))


def get_num_flop_per_token(
    num_non_embed_params: int, n_layers: int, dim: int, seq_len: int
) -> int:
    return 6 * num_non_embed_params + attention_flops_per_token(
        n_layers, seq_len, dim, True
    )


def causal_mask(b, h, q_idx, kv_idx):
    return q_idx >= kv_idx


@dataclass
class LMTransformerArgs(BaseTransformerArgs):

    seed: int = 42

    vocab_size: int = -1
    weight_tying: bool = False

    sliding_window: Optional[int] = None

    freeze_language_model: Optional[bool] = False
    freeze_vision_model: Optional[bool] = False

    vision_model: Optional[Dict[str, Any]] = None

    mlp_init: InitArgs = field(default_factory=InitArgs)
    pooling_ratio: int = 1
    remove_vision_class_token: bool = True

    attn_impl: str = "sdpa"


class LMTransformer(BaseTransformer):
    def __init__(self, args: LMTransformerArgs):
        super().__init__(args)
        self.weight_tying = args.weight_tying
        self.sliding_window = args.sliding_window

        assert args.vocab_size > 0

        self.tok_embeddings = torch.nn.Embedding(args.vocab_size, args.dim)

        self.norm = RMSNorm(args.dim, eps=args.norm_eps)

        self.output = nn.Linear(
            args.dim,
            args.vocab_size,
            bias=False,
        )

        if args.weight_tying:
            self.output = TiedLinear(self.tok_embeddings)
        else:
            self.output = nn.Linear(
                args.dim,
                args.vocab_size,
                bias=False,
            )

        if args.vision_model:
            logger.info(
                f"Initializing PE_VisionTransformer with args: {args.vision_model}"
            )
            self.vision_model = PE_VisionTransformer(**args.vision_model, output_dim=None)
            self.vision_projector = MLPProjector(args)

        self.freeze_vision_model = args.freeze_vision_model
        self.freeze_language_model = args.freeze_language_model

    def train(self, mode: bool = True):
        super().train(mode=mode)
        for name, param in self.named_parameters():
            if "vision_model" in name:
                param.requires_grad = mode and not self.freeze_vision_model
            elif "vision_projector" in name:
                param.requires_grad = mode
            else:
                param.requires_grad = mode and not self.freeze_language_model
        return self

    def forward(
        self,
        token_values: torch.Tensor,
        target: Optional[torch.Tensor] = None,
        tok_idx: Optional[torch.Tensor] = None,
        mask: Optional[Union[BlockMask, AttentionBias, torch.Tensor, str]] = None,
        images: Optional[torch.Tensor] = None,
        image_pos_index: Optional[torch.Tensor] = None,
        loss_mask: Optional[torch.Tensor] = None,
        aspect_ratios: Optional[torch.Tensor] = None,
        num_chunks: List[int] = [1],
        media_type: List[str] = ["multi_image"],
        attn_impl: str = "sdpa",
    ):
        _, seqlen = token_values.shape

        h = self.tok_embeddings(token_values)

        if images is not None:
            h_img = self.vision_model(images, strip_cls_token=True)
            h_img = self.vision_projector(h_img)

            h = self.stitch_images_into_text(
                h,
                h_img,
                image_pos_index,
                num_chunks=num_chunks,
                media_type=media_type,
            )

        mask = (
            mask
            if mask is not None
            else create_causal_mask(seqlen, attn_impl, self.sliding_window)
        )

        h = super().forward(h, tok_idx=tok_idx, mask=mask, attn_impl=attn_impl)

        logits = self.output(self.norm(h))
        if target is not None:
            logits = logits[loss_mask]
            target = target[loss_mask]
            return cross_entropy(logits, target)
        else:
            return logits

    def reset_parameters(self, init_std=None):
        # Either use fixed base std or sqrt model dim
        super().reset_parameters()
        init_std = init_std or (self.dim ** (-0.5))
        self.norm.reset_parameters()
        nn.init.trunc_normal_(
            self.tok_embeddings.weight,
            mean=0.0,
            std=init_std,
            a=-3 * init_std,
            b=3 * init_std,
        )
        if not self.weight_tying:
            nn.init.trunc_normal_(
                self.output.weight,
                mean=0.0,
                std=init_std,
                a=-3 * init_std,
                b=3 * init_std,
            )

    def stitch_images_into_text(
        self,
        h_tok: torch.Tensor,
        h_img: List[torch.Tensor],
        image_pos_index: torch.Tensor,
        num_chunks: List[int],
        media_type: List[str],
    ):
        # Generate cumulative indices for each sample
        cumulative_indices = list(itertools.accumulate(num_chunks, initial=0))
        # Get indices for non-text samples
        non_text_indices = [
            idx
            for start, end, m_type in zip(
                cumulative_indices[:-1], cumulative_indices[1:], media_type
            )
            if m_type != "text"
            for idx in range(start, end)
        ]
        img_indices_B, img_indices_L = torch.where(image_pos_index >= 0)
        valid_index_filter = img_indices_L < h_tok.shape[1]
        img_indices_L = img_indices_L[valid_index_filter]
        img_indices_B = img_indices_B[valid_index_filter]
        h_tok[img_indices_B, img_indices_L] = h_img[non_text_indices].flatten(0, 1)[
            valid_index_filter
        ]
        return h_tok


# Optional policy for activation checkpointing. With None, we stick to the default (defined distributed.py: default_no_recompute_ops)
def get_no_recompute_ops():
    return None


# Optional and only used for fully shard options (fsdp) is choose. Highly recommanded for large models
def build_fsdp_grouping_plan(model_args: LMTransformerArgs):
    group_plan: Tuple[int, bool] = []

    # Grouping and output seperately
    group_plan.append(("tok_embeddings", False))

    group_plan.append(("vision_model", False))
    group_plan.append(("vision_projector", False))

    # Grouping by layers
    for i in range(model_args.n_layers):
        group_plan.append((f"layers.{i}", True))

    group_plan.append(("output", True))

    return group_plan


# Optional and only used for model/tensor parallelism when tp_size > 1
def tp_parallelize(model, tp_mesh, model_args: LMTransformerArgs, distributed_args):
    assert model_args.dim % distributed_args.tp_size == 0
    assert model_args.vocab_size % distributed_args.tp_size == 0
    assert model_args.n_heads % distributed_args.tp_size == 0
    assert (model_args.n_kv_heads or 0) % distributed_args.tp_size == 0
    assert model_args.n_heads % (model_args.n_kv_heads or 1) == 0

    # Embedding layer tp
    main_plan = {}
    main_plan["tok_embeddings"] = ColwiseParallel(
        input_layouts=Replicate(), output_layouts=Shard(1)
    )
    main_plan["norm"] = SequenceParallel()
    main_plan["output"] = ColwiseParallel(
        input_layouts=Shard(1), output_layouts=Replicate()
    )

    parallelize_module(
        model,
        tp_mesh,
        main_plan,
    )

    # Attention layers tp
    for layer in model.layers:
        layer_plan = {}

        layer_plan["attention"] = PrepareModuleInput(
            input_layouts=(Shard(1), None),
            desired_input_layouts=(Replicate(), None),
        )
        layer_plan["attention_norm"] = SequenceParallel()
        layer_plan["attention.wq"] = ColwiseParallel()
        layer_plan["attention.wk"] = ColwiseParallel()
        layer_plan["attention.wv"] = ColwiseParallel()
        layer_plan["attention.wo"] = RowwiseParallel(output_layouts=Shard(1))

        # Feedforward layers tp
        layer_plan["feed_forward"] = PrepareModuleInput(
            input_layouts=(Shard(1),),
            desired_input_layouts=(Replicate(),),
        )
        layer_plan["ffn_norm"] = SequenceParallel()
        layer_plan["feed_forward.w1"] = ColwiseParallel()
        layer_plan["feed_forward.w3"] = ColwiseParallel()
        layer_plan["feed_forward.w2"] = RowwiseParallel(output_layouts=Shard(1))

        parallelize_module(
            layer,
            tp_mesh,
            layer_plan,
        )

        # Adjusting the number of heads and kv heads according to the tp size
        attn_layer = layer.attention
        attn_layer.n_heads = attn_layer.n_heads // distributed_args.tp_size
        attn_layer.n_kv_heads = attn_layer.n_kv_heads // distributed_args.tp_size



#    Copyright 2023 Haotian Liu
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from abc import ABC, abstractmethod

import torch
import torch.nn as nn
from PIL import Image
from io import BytesIO
import base64
import torch
import math
import ast

from .multimodal_encoder.builder import build_vision_tower
from .multimodal_projector.builder import build_vision_projector

# Model Constants
IGNORE_INDEX = -100
IMAGE_TOKEN_INDEX = -200
DEFAULT_IMAGE_TOKEN = "<image>"
DEFAULT_IMAGE_PATCH_TOKEN = "<im_patch>"
DEFAULT_IM_START_TOKEN = "<im_start>"
DEFAULT_IM_END_TOKEN = "<im_end>"

def select_best_resolution(original_size, possible_resolutions):
    """
    Selects the best resolution from a list of possible resolutions based on the original size.

    Args:
        original_size (tuple): The original size of the image in the format (width, height).
        possible_resolutions (list): A list of possible resolutions in the format [(width1, height1), (width2, height2), ...].

    Returns:
        tuple: The best fit resolution in the format (width, height).
    """
    original_width, original_height = original_size
    best_fit = None
    max_effective_resolution = 0
    min_wasted_resolution = float('inf')

    for width, height in possible_resolutions:
        scale = min(width / original_width, height / original_height)
        downscaled_width, downscaled_height = int(original_width * scale), int(original_height * scale)
        effective_resolution = min(downscaled_width * downscaled_height, original_width * original_height)
        wasted_resolution = (width * height) - effective_resolution

        if effective_resolution > max_effective_resolution or (effective_resolution == max_effective_resolution and wasted_resolution < min_wasted_resolution):
            max_effective_resolution = effective_resolution
            min_wasted_resolution = wasted_resolution
            best_fit = (width, height)

    return best_fit


def resize_and_pad_image(image, target_resolution):
    """
    Resize and pad an image to a target resolution while maintaining aspect ratio.

    Args:
        image (PIL.Image.Image): The input image.
        target_resolution (tuple): The target resolution (width, height) of the image.

    Returns:
        PIL.Image.Image: The resized and padded image.
    """
    original_width, original_height = image.size
    target_width, target_height = target_resolution

    scale_w = target_width / original_width
    scale_h = target_height / original_height

    if scale_w < scale_h:
        new_width = target_width
        new_height = min(math.ceil(original_height * scale_w), target_height)
    else:
        new_height = target_height
        new_width = min(math.ceil(original_width * scale_h), target_width)

    # Resize the image
    resized_image = image.resize((new_width, new_height))

    new_image = Image.new('RGB', (target_width, target_height), (0, 0, 0))
    paste_x = (target_width - new_width) // 2
    paste_y = (target_height - new_height) // 2
    new_image.paste(resized_image, (paste_x, paste_y))

    return new_image


def divide_to_patches(image, patch_size):
    """
    Divides an image into patches of a specified size.

    Args:
        image (PIL.Image.Image): The input image.
        patch_size (int): The size of each patch.

    Returns:
        list: A list of PIL.Image.Image objects representing the patches.
    """
    patches = []
    width, height = image.size
    for i in range(0, height, patch_size):
        for j in range(0, width, patch_size):
            box = (j, i, j + patch_size, i + patch_size)
            patch = image.crop(box)
            patches.append(patch)

    return patches


def get_anyres_image_grid_shape(image_size, grid_pinpoints, patch_size):
    """
    Calculate the shape of the image patch grid after the preprocessing for images of any resolution.

    Args:
        image_size (tuple): The size of the input image in the format (width, height).
        grid_pinpoints (str): A string representation of a list of possible resolutions.
        patch_size (int): The size of each image patch.

    Returns:
        tuple: The shape of the image patch grid in the format (width, height).
    """
    if type(grid_pinpoints) is list:
        possible_resolutions = grid_pinpoints
    else:
        possible_resolutions = ast.literal_eval(grid_pinpoints)
    width, height = select_best_resolution(image_size, possible_resolutions)
    return width // patch_size, height // patch_size


class LlavaMetaModel:

    def __init__(self, config):
        super(LlavaMetaModel, self).__init__(config)

        if hasattr(config, "mm_vision_tower"):
            self.vision_tower = build_vision_tower(config, delay_load=True)
            self.mm_projector = build_vision_projector(config)

            if 'unpad' in getattr(config, 'mm_patch_merge_type', ''):
                self.image_newline = nn.Parameter(
                    torch.empty(config.hidden_size, dtype=self.dtype)
                )

    def get_vision_tower(self):
        vision_tower = getattr(self, 'vision_tower', None)
        if type(vision_tower) is list:
            vision_tower = vision_tower[0]
        return vision_tower

    def initialize_vision_modules(self, model_args, fsdp=None):
        vision_tower = model_args.vision_tower
        mm_vision_select_layer = model_args.mm_vision_select_layer
        mm_vision_select_feature = model_args.mm_vision_select_feature
        pretrain_mm_mlp_adapter = model_args.pretrain_mm_mlp_adapter
        mm_patch_merge_type = model_args.mm_patch_merge_type

        self.config.mm_vision_tower = vision_tower

        if self.get_vision_tower() is None:
            vision_tower = build_vision_tower(model_args)

            if fsdp is not None and len(fsdp) > 0:
                self.vision_tower = [vision_tower]
            else:
                self.vision_tower = vision_tower
        else:
            if fsdp is not None and len(fsdp) > 0:
                vision_tower = self.vision_tower[0]
            else:
                vision_tower = self.vision_tower
            vision_tower.load_model()

        self.config.use_mm_proj = True
        self.config.mm_projector_type = getattr(model_args, 'mm_projector_type', 'linear')
        self.config.mm_hidden_size = vision_tower.hidden_size
        self.config.mm_vision_select_layer = mm_vision_select_layer
        self.config.mm_vision_select_feature = mm_vision_select_feature
        self.config.mm_patch_merge_type = mm_patch_merge_type

        if getattr(self, 'mm_projector', None) is None:
            self.mm_projector = build_vision_projector(self.config)

            if 'unpad' in mm_patch_merge_type:
                embed_std = 1 / torch.sqrt(torch.tensor(self.config.hidden_size, dtype=self.dtype))
                self.image_newline = nn.Parameter(
                    torch.randn(self.config.hidden_size, dtype=self.dtype) * embed_std
                )
        else:
            # In case it is frozen by LoRA
            for p in self.mm_projector.parameters():
                p.requires_grad = True

        if pretrain_mm_mlp_adapter is not None:
            mm_projector_weights = torch.load(pretrain_mm_mlp_adapter, map_location='cpu')
            def get_w(weights, keyword):
                return {k.split(keyword + '.')[1]: v for k, v in weights.items() if keyword in k}

            self.mm_projector.load_state_dict(get_w(mm_projector_weights, 'mm_projector'))


def unpad_image(tensor, original_size):
    """
    Unpads a PyTorch tensor of a padded and resized image.

    Args:
    tensor (torch.Tensor): The image tensor, assumed to be in CxHxW format.
    original_size (tuple): The original size of the image (height, width).

    Returns:
    torch.Tensor: The unpadded image tensor.
    """
    original_width, original_height = original_size
    current_height, current_width = tensor.shape[1:]

    original_aspect_ratio = original_width / original_height
    current_aspect_ratio = current_width / current_height

    if original_aspect_ratio > current_aspect_ratio:
        scale_factor = current_width / original_width
        new_height = int(original_height * scale_factor)
        padding = (current_height - new_height) // 2
        unpadded_tensor = tensor[:, padding:current_height - padding, :]
    else:
        scale_factor = current_height / original_height
        new_width = int(original_width * scale_factor)
        padding = (current_width - new_width) // 2
        unpadded_tensor = tensor[:, :, padding:current_width - padding]

    return unpadded_tensor


class LlavaMetaForCausalLM(ABC):

    @abstractmethod
    def get_model(self):
        pass

    def get_vision_tower(self):
        return self.get_model().get_vision_tower()

    def encode_images(self, images):
        image_features = self.get_model().get_vision_tower()(images)
        image_features = self.get_model().mm_projector(image_features)
        return image_features

    def prepare_inputs_labels_for_multimodal(
        self, input_ids, position_ids, attention_mask, past_key_values, labels,
        images, image_sizes=None
    ):
        vision_tower = self.get_vision_tower()
        if vision_tower is None or images is None or input_ids.shape[1] == 1:
            return input_ids, position_ids, attention_mask, past_key_values, None, labels

        if type(images) is list or images.ndim == 5:
            if type(images) is list:
                images = [x.unsqueeze(0) if x.ndim == 3 else x for x in images]
            concat_images = torch.cat([image for image in images], dim=0)
            image_features = self.encode_images(concat_images)
            split_sizes = [image.shape[0] for image in images]
            image_features = torch.split(image_features, split_sizes, dim=0)
            mm_patch_merge_type = getattr(self.config, 'mm_patch_merge_type', 'flat')
            image_aspect_ratio = getattr(self.config, 'image_aspect_ratio', 'square')
            if mm_patch_merge_type == 'flat':
                image_features = [x.flatten(0, 1) for x in image_features]
            elif mm_patch_merge_type.startswith('spatial'):
                new_image_features = []
                for image_idx, image_feature in enumerate(image_features):
                    if image_feature.shape[0] > 1:
                        base_image_feature = image_feature[0]
                        image_feature = image_feature[1:]
                        height = width = self.get_vision_tower().num_patches_per_side
                        assert height * width == base_image_feature.shape[0]
                        if image_aspect_ratio == 'anyres':
                            num_patch_width, num_patch_height = get_anyres_image_grid_shape(image_sizes[image_idx], self.config.image_grid_pinpoints, self.get_vision_tower().config.image_size)
                            image_feature = image_feature.view(num_patch_height, num_patch_width, height, width, -1)
                        else:
                            raise NotImplementedError
                        if 'unpad' in mm_patch_merge_type:
                            image_feature = image_feature.permute(4, 0, 2, 1, 3).contiguous()
                            image_feature = image_feature.flatten(1, 2).flatten(2, 3)
                            image_feature = unpad_image(image_feature, image_sizes[image_idx])
                            image_feature = torch.cat((
                                image_feature,
                                self.model.image_newline[:, None, None].expand(*image_feature.shape[:-1], 1)
                            ), dim=-1)
                            image_feature = image_feature.flatten(1, 2).transpose(0, 1)
                        else:
                            image_feature = image_feature.permute(0, 2, 1, 3, 4).contiguous()
                            image_feature = image_feature.flatten(0, 3)
                        image_feature = torch.cat((base_image_feature, image_feature), dim=0)
                    else:
                        image_feature = image_feature[0]
                        if 'unpad' in mm_patch_merge_type:
                            image_feature = torch.cat((
                                image_feature,
                                self.model.image_newline[None]
                            ), dim=0)
                    new_image_features.append(image_feature)
                image_features = new_image_features
            else:
                raise ValueError(f"Unexpected mm_patch_merge_type: {self.config.mm_patch_merge_type}")
        else:
            image_features = self.encode_images(images)

        # TODO: image start / end is not implemented here to support pretraining.
        if getattr(self.config, 'tune_mm_mlp_adapter', False) and getattr(self.config, 'mm_use_im_start_end', False):
            raise NotImplementedError

        # Let's just add dummy tensors if they do not exist,
        # it is a headache to deal with None all the time.
        # But it is not ideal, and if you have a better idea,
        # please open an issue / submit a PR, thanks.
        _labels = labels
        _position_ids = position_ids
        _attention_mask = attention_mask
        if attention_mask is None:
            attention_mask = torch.ones_like(input_ids, dtype=torch.bool)
        else:
            attention_mask = attention_mask.bool()
        if position_ids is None:
            position_ids = torch.arange(0, input_ids.shape[1], dtype=torch.long, device=input_ids.device)
        if labels is None:
            labels = torch.full_like(input_ids, IGNORE_INDEX)

        # remove the padding using attention_mask -- FIXME
        _input_ids = input_ids
        input_ids = [cur_input_ids[cur_attention_mask] for cur_input_ids, cur_attention_mask in zip(input_ids, attention_mask)]
        labels = [cur_labels[cur_attention_mask] for cur_labels, cur_attention_mask in zip(labels, attention_mask)]

        new_input_embeds = []
        new_labels = []
        cur_image_idx = 0
        for batch_idx, cur_input_ids in enumerate(input_ids):
            num_images = (cur_input_ids == IMAGE_TOKEN_INDEX).sum()
            if num_images == 0:
                cur_image_features = image_features[cur_image_idx]
                cur_input_embeds_1 = self.get_model().embed_tokens(cur_input_ids)
                cur_input_embeds = torch.cat([cur_input_embeds_1, cur_image_features[0:0]], dim=0)
                new_input_embeds.append(cur_input_embeds)
                new_labels.append(labels[batch_idx])
                cur_image_idx += 1
                continue

            image_token_indices = [-1] + torch.where(cur_input_ids == IMAGE_TOKEN_INDEX)[0].tolist() + [cur_input_ids.shape[0]]
            cur_input_ids_noim = []
            cur_labels = labels[batch_idx]
            cur_labels_noim = []
            for i in range(len(image_token_indices) - 1):
                cur_input_ids_noim.append(cur_input_ids[image_token_indices[i]+1:image_token_indices[i+1]])
                cur_labels_noim.append(cur_labels[image_token_indices[i]+1:image_token_indices[i+1]])
            split_sizes = [x.shape[0] for x in cur_labels_noim]
            cur_input_embeds = self.get_model().embed_tokens(torch.cat(cur_input_ids_noim))
            cur_input_embeds_no_im = torch.split(cur_input_embeds, split_sizes, dim=0)
            cur_new_input_embeds = []
            cur_new_labels = []

            for i in range(num_images + 1):
                cur_new_input_embeds.append(cur_input_embeds_no_im[i])
                cur_new_labels.append(cur_labels_noim[i])
                if i < num_images:
                    cur_image_features = image_features[cur_image_idx]
                    cur_image_idx += 1
                    cur_new_input_embeds.append(cur_image_features)
                    cur_new_labels.append(torch.full((cur_image_features.shape[0],), IGNORE_INDEX, device=cur_labels.device, dtype=cur_labels.dtype))

            cur_new_input_embeds = [x.to(self.device) for x in cur_new_input_embeds]

            cur_new_input_embeds = torch.cat(cur_new_input_embeds)
            cur_new_labels = torch.cat(cur_new_labels)

            new_input_embeds.append(cur_new_input_embeds)
            new_labels.append(cur_new_labels)

        # Truncate sequences to max length as image embeddings can make the sequence longer
        tokenizer_model_max_length = getattr(self.config, 'tokenizer_model_max_length', None)
        if tokenizer_model_max_length is not None:
            new_input_embeds = [x[:tokenizer_model_max_length] for x in new_input_embeds]
            new_labels = [x[:tokenizer_model_max_length] for x in new_labels]

        # Combine them
        max_len = max(x.shape[0] for x in new_input_embeds)
        batch_size = len(new_input_embeds)

        new_input_embeds_padded = []
        new_labels_padded = torch.full((batch_size, max_len), IGNORE_INDEX, dtype=new_labels[0].dtype, device=new_labels[0].device)
        attention_mask = torch.zeros((batch_size, max_len), dtype=attention_mask.dtype, device=attention_mask.device)
        position_ids = torch.zeros((batch_size, max_len), dtype=position_ids.dtype, device=position_ids.device)

        for i, (cur_new_embed, cur_new_labels) in enumerate(zip(new_input_embeds, new_labels)):
            cur_len = cur_new_embed.shape[0]
            if getattr(self.config, 'tokenizer_padding_side', 'right') == "left":
                new_input_embeds_padded.append(torch.cat((
                    torch.zeros((max_len - cur_len, cur_new_embed.shape[1]), dtype=cur_new_embed.dtype, device=cur_new_embed.device),
                    cur_new_embed
                ), dim=0))
                if cur_len > 0:
                    new_labels_padded[i, -cur_len:] = cur_new_labels
                    attention_mask[i, -cur_len:] = True
                    position_ids[i, -cur_len:] = torch.arange(0, cur_len, dtype=position_ids.dtype, device=position_ids.device)
            else:
                new_input_embeds_padded.append(torch.cat((
                    cur_new_embed,
                    torch.zeros((max_len - cur_len, cur_new_embed.shape[1]), dtype=cur_new_embed.dtype, device=cur_new_embed.device)
                ), dim=0))
                if cur_len > 0:
                    new_labels_padded[i, :cur_len] = cur_new_labels
                    attention_mask[i, :cur_len] = True
                    position_ids[i, :cur_len] = torch.arange(0, cur_len, dtype=position_ids.dtype, device=position_ids.device)

        new_input_embeds = torch.stack(new_input_embeds_padded, dim=0)

        if _labels is None:
            new_labels = None
        else:
            new_labels = new_labels_padded

        if _attention_mask is None:
            attention_mask = None
        else:
            attention_mask = attention_mask.to(dtype=_attention_mask.dtype)

        if _position_ids is None:
            position_ids = None

        return None, position_ids, attention_mask, past_key_values, new_input_embeds, new_labels

    def initialize_vision_tokenizer(self, model_args, tokenizer):
        if model_args.mm_use_im_patch_token:
            tokenizer.add_tokens([DEFAULT_IMAGE_PATCH_TOKEN], special_tokens=True)
            self.resize_token_embeddings(len(tokenizer))

        if model_args.mm_use_im_start_end:
            num_new_tokens = tokenizer.add_tokens([DEFAULT_IM_START_TOKEN, DEFAULT_IM_END_TOKEN], special_tokens=True)
            self.resize_token_embeddings(len(tokenizer))

            if num_new_tokens > 0:
                input_embeddings = self.get_input_embeddings().weight.data
                output_embeddings = self.get_output_embeddings().weight.data

                input_embeddings_avg = input_embeddings[:-num_new_tokens].mean(
                    dim=0, keepdim=True)
                output_embeddings_avg = output_embeddings[:-num_new_tokens].mean(
                    dim=0, keepdim=True)

                input_embeddings[-num_new_tokens:] = input_embeddings_avg
                output_embeddings[-num_new_tokens:] = output_embeddings_avg

            if model_args.tune_mm_mlp_adapter:
                for p in self.get_input_embeddings().parameters():
                    p.requires_grad = True
                for p in self.get_output_embeddings().parameters():
                    p.requires_grad = False

            if model_args.pretrain_mm_mlp_adapter:
                mm_projector_weights = torch.load(model_args.pretrain_mm_mlp_adapter, map_location='cpu')
                embed_tokens_weight = mm_projector_weights['model.embed_tokens.weight']
                assert num_new_tokens == 2
                if input_embeddings.shape == embed_tokens_weight.shape:
                    input_embeddings[-num_new_tokens:] = embed_tokens_weight[-num_new_tokens:]
                elif embed_tokens_weight.shape[0] == num_new_tokens:
                    input_embeddings[-num_new_tokens:] = embed_tokens_weight
                else:
                    raise ValueError(f"Unexpected embed_tokens_weight shape. Pretrained: {embed_tokens_weight.shape}. Current: {input_embeddings.shape}. Numer of new tokens: {num_new_tokens}.")
        elif model_args.mm_use_im_patch_token:
            if model_args.tune_mm_mlp_adapter:
                for p in self.get_input_embeddings().parameters():
                    p.requires_grad = False
                for p in self.get_output_embeddings().parameters():
                    p.requires_grad = False


# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
COCO dataset which returns image_id for evaluation.

Mostly copy-paste from https://github.com/pytorch/vision/blob/13b35ff/references/detection/coco_utils.py
"""
import random
from pathlib import Path

import datasets.transforms as T
import torch
import torch.utils.data
import torchvision.transforms.functional as F
from pycocotools import mask as coco_mask
from util.misc import get_local_rank, get_local_size

from .torchvision_datasets import CocoDetection as TvCocoDetection



class CocoDetection(TvCocoDetection):
    def __init__(
        self,
        img_folder,
        ann_file,
        transforms,
        return_masks,
        cache_mode=False,
        local_rank=0,
        local_size=1,
        test_hflip_aug=False,
        tta=False,
        is_train=False,
        lsj_img_size=1824,
    ):
        super(CocoDetection, self).__init__(
            img_folder,
            ann_file,
            cache_mode=cache_mode,
            local_rank=local_rank,
            local_size=local_size,
        )
        self._transforms = transforms
        self.prepare = ConvertCocoPolysToMask(return_masks)
        self.test_hflip_aug = test_hflip_aug
        self.tta = tta
        if lsj_img_size == 1728: # for back-compatibility
            self.tta_image_size = [1536, 1152,]
        else:
            self.tta_image_size = [1728, 1536, 1344,]
        
        self.is_train = is_train

    def __getitem__(self, idx):
        img, target = super(CocoDetection, self).__getitem__(idx)
        image_id = self.ids[idx]
        target = {"image_id": image_id, "annotations": target}
        img, target = self.prepare(img, target)
        if self._transforms is not None:
            img, target = self._transforms(img, target)

        if self.test_hflip_aug:
            flipped_img = torch.flip(img, dims=[-1])
            new_img = torch.cat([img, flipped_img], dim=0)
            return new_img, target

        elif self.tta:
            tta_images = [img]
            flipped_img = torch.flip(img, dims=[-1])
            tta_images.append(flipped_img)
            _, height, width = img.shape
            max_size_len = height if height >= width else width
            for new_max_size in self.tta_image_size:
                scale = new_max_size / max_size_len
                new_height, new_width = int(scale * height), int(scale * width)
                new_img = F.resize(img, size=(new_height, new_width))
                tta_images.append(new_img)
                flipped_img = torch.flip(new_img, dims=[-1])
                tta_images.append(flipped_img)
            return tta_images, target
        else:
            return img, target


def convert_coco_poly_to_mask(segmentations, height, width):
    masks = []
    for polygons in segmentations:
        rles = coco_mask.frPyObjects(polygons, height, width)
        mask = coco_mask.decode(rles)
        if len(mask.shape) < 3:
            mask = mask[..., None]
        mask = torch.as_tensor(mask, dtype=torch.uint8)
        mask = mask.any(dim=2)
        masks.append(mask)
    if masks:
        masks = torch.stack(masks, dim=0)
    else:
        masks = torch.zeros((0, height, width), dtype=torch.uint8)
    return masks


class ConvertCocoPolysToMask(object):
    def __init__(self, return_masks=False):
        self.return_masks = return_masks

    def __call__(self, image, target):
        w, h = image.size

        image_id = target["image_id"]
        image_id = torch.tensor([image_id])

        anno = target["annotations"]

        anno = [obj for obj in anno if "iscrowd" not in obj or obj["iscrowd"] == 0]

        boxes = [obj["bbox"] for obj in anno]
        # guard against no boxes via resizing
        boxes = torch.as_tensor(boxes, dtype=torch.float32).reshape(-1, 4)
        boxes[:, 2:] += boxes[:, :2]
        boxes[:, 0::2].clamp_(min=0, max=w)
        boxes[:, 1::2].clamp_(min=0, max=h)

        classes = [obj["category_id"] for obj in anno]
        classes = torch.tensor(classes, dtype=torch.int64)

        if self.return_masks:
            segmentations = [obj["segmentation"] for obj in anno]
            masks = convert_coco_poly_to_mask(segmentations, h, w)

        keypoints = None
        if anno and "keypoints" in anno[0]:
            keypoints = [obj["keypoints"] for obj in anno]
            keypoints = torch.as_tensor(keypoints, dtype=torch.float32)
            num_keypoints = keypoints.shape[0]
            if num_keypoints:
                keypoints = keypoints.view(num_keypoints, -1, 3)

        keep = (boxes[:, 3] > boxes[:, 1]) & (boxes[:, 2] > boxes[:, 0])
        boxes = boxes[keep]
        classes = classes[keep]
        if self.return_masks:
            masks = masks[keep]
        if keypoints is not None:
            keypoints = keypoints[keep]

        target = {}
        target["boxes"] = boxes
        target["labels"] = classes
        if self.return_masks:
            target["masks"] = masks
        target["image_id"] = image_id
        if keypoints is not None:
            target["keypoints"] = keypoints

        # for conversion to coco api
        area = torch.tensor([obj["area"] for obj in anno])
        iscrowd = torch.tensor(
            [obj["iscrowd"] if "iscrowd" in obj else 0 for obj in anno]
        )
        target["area"] = area[keep]
        target["iscrowd"] = iscrowd[keep]

        target["orig_size"] = torch.as_tensor([int(h), int(w)])
        target["size"] = torch.as_tensor([int(h), int(w)])

        return image, target


def make_coco_transforms(image_set, bigger):

    normalize = T.Compose(
        [T.ToTensor(), T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])]
    )

    if "train" in image_set:
        scales = [480, 512, 544, 576, 608, 640, 672, 704, 736, 768, 800]
    if "val" in image_set or "test" in image_set:
        scales = [800]

    max_size = 1333
    if bigger:
        scales = [int(1.5 * s) for s in scales]
        max_size = 2000

    if image_set == "train":
        augmentation_list = [
            T.RandomHorizontalFlip(),
            T.RandomSelect(
                T.RandomResize(scales, max_size=max_size),
                T.Compose(
                    [
                        T.RandomResize([400, 500, 600]),
                        T.RandomSizeCrop(384, 600),
                        T.RandomResize(scales, max_size=max_size),
                    ]
                ),
            ),
            normalize,
        ]

        return T.Compose(augmentation_list)

    if image_set == "val":
        return T.Compose(
            [
                T.RandomResize(scales, max_size=max_size),
                normalize,
            ]
        )

    raise ValueError(f"unknown {image_set}")


def make_coco_transforms_lsj(
    image_set, image_size, lsj_img_train_min=480, lsj_strong_aug=False
):
    """
    Reference: https://github.com/facebookresearch/detectron2/blob/main/projects/ViTDet/configs/common/coco_loader_lsj.py

    import detectron2.data.transforms as T
    from detectron2 import model_zoo
    from detectron2.config import LazyCall as L

    # Data using LSJ
    image_size = 1024
    dataloader = model_zoo.get_config("common/data/coco.py").dataloader
    dataloader.train.mapper.augmentations = [
        L(T.RandomFlip)(horizontal=True),  # flip first
        L(T.ResizeScale)(
            min_scale=0.1, max_scale=2.0, target_height=image_size, target_width=image_size
        ),
        L(T.FixedSizeCrop)(crop_size=(image_size, image_size), pad=False),
    ]
    dataloader.train.mapper.image_format = "RGB"
    dataloader.train.total_batch_size = 64
    # recompute boxes due to cropping
    dataloader.train.mapper.recompute_boxes = True

    dataloader.test.mapper.augmentations = [
        L(T.ResizeShortestEdge)(short_edge_length=image_size, max_size=image_size),
    ]
    """

    """
    In our implementation, we simulate lsj data augmentation by:
    (1) first the following augmentations 
    (2) then padding to (image_size, image_size) in collator, see util/misc/collate_fn_lsj.py
    """
    normalize = T.Compose(
        [T.ToTensor(), T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])]
    )

    if "train" in image_set:
        scales = [scale for scale in range(lsj_img_train_min, image_size, 32)]
    if "val" in image_set or "test" in image_set or "unlabel" in image_set:
        scales = [image_size - 32]

    # max_size = 1333
    # if bigger:
    #     scales = [int(1.5 * s) for s in scales]
    #     max_size = 2000
    max_size = image_size - 32  # for some wired bugs

    augmentation_list = []
    if "train" in image_set:
        if lsj_strong_aug:
            augmentation_list.extend(
                [
                    T.ColorJitter((0.4, 0.4, 0.4, 0.1), p=0.5),
                    T.RandomGrayscale(p=0.2),
                    # T.RandomErasingP05(),
                ]
            )
        augmentation_list.extend(
            [
                T.RandomHorizontalFlip(),
                T.RandomSelect(
                    # similar to (T.ResizeScale)(min_scale=0.1, max_scale=1.0, target_height=image_size, target_width=image_size) and pad
                    T.RandomResize(scales, max_size=max_size),
                    # similar to (T.ResizeScale)(min_scale=1.0, max_scale=2.0, target_height=image_size, target_width=image_size) and crop
                    T.Compose(
                        [
                            T.RandomResize([400, 500, 600]),
                            T.RandomSizeCrop(384, 600),
                            T.RandomResize([max_size], max_size=max_size),
                        ]
                    ),
                ),
                normalize,
            ]
        )
        return T.Compose(augmentation_list)

    if image_set == "val":
        return T.Compose(
            [
                T.RandomResize(scales, max_size=max_size),
                normalize,
            ]
        )

    raise ValueError(f"unknown {image_set}")


def build(image_set, args):
    root = Path(args.coco_path)
    assert root.exists(), f"provided COCO path {root} does not exist"
    mode = "instances"
    PATHS = {
        "train": (root / "train2017", root / "annotations" / f"{mode}_train2017.json"),
        "val": (root / "val2017", root / "annotations" / f"{mode}_val2017.json"),
    }

    img_folder, ann_file = PATHS[image_set]
    if args.lsj:
        coco_transform = make_coco_transforms_lsj(
            image_set,
            args.lsj_img_size,
            args.lsj_img_train_min,
            args.lsj_strong_aug,
        )
    else:
        coco_transform = make_coco_transforms(image_set, args.bigger)
    dataset = CocoDetection(
        img_folder,
        ann_file,
        transforms=coco_transform,
        return_masks=args.masks,
        cache_mode=args.cache_mode,
        local_rank=get_local_rank(),
        local_size=get_local_size(),
        test_hflip_aug=args.test_hflip_aug,
        tta=args.tta,
        is_train=("train" in image_set),
        lsj_img_size=args.lsj_img_size,
    )
    return dataset
